; File name: EECounter.asm
; Last Update: May 22, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Program to demonstrate on chip EEPROM data memory read
; and write operation. Program uses LCD display to output
; results.
; Operation:
; The program keeps track and displays the inNum of times
; the code has been started.
; For LCD display parameters see the LCDTest2 program. 
; WARNING:
; Code assumes 4Mhz clock. Delay routines must be
; edited for faster clock
;
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

;=========================
; setup and configuration
;=========================
	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF

;=====================================================
;                 constant definitions
;  for PIC-to-LCD pin wiring and LCD line addresses
;=====================================================
#define E_line 1   	;|
#define RS_line 2	;| -- from wiring diagram 
#define RW_line 3	;|
; LCD line addresses (from LCD data sheet)
#define LCD_1 0x80	; First LCD line constant
#define LCD_2 0xc0	; Second LCD line constant
; Note: The constant that define the LCD display line
;       addresses have the high-order bit set in
;       order to faciliate the controller command 
;
;=====================================================
;              variables in PIC RAM
;=====================================================
; Reserve 16 bytes for string buffer
	cblock	0x0c
	strData
	endc
; Reserve three bytes for ASCII digits
	cblock	0x1d
	asc100
	asc10
	asc1 
	endc
; Continue with local variables
	cblock	0x20		; Start of block
	count1		; Counter # 1
	count2		; Counter # 2
	count3		; Counter # 3
	pic_ad		; Storage for start of text area
				; (labeled strData) in PIC RAM
	J			; counter J
	K			; counter K
	index		; Index into text table (also used
				; for auxiliary storage)
	store1		; Local temporary storage 
	store2		; Storage # 2
; EEPROM-related variables
	EEMemAdd	; EEPROM address to access
	EEByte		; Data byte to write
; Storage for ASCII decimal conversion and digits
	inNum		; Source operand
	thisDig		; Digit counter
	endc

;=========================================================
;                           program
;=========================================================
		org		0	  ; start at address 
		goto	main
; Space for interrupt handlers
	org		0x08

main:
	movlw	b'00000000' ; All lines to output
	tris	PORTA		; in port A
	tris	PORTB		; and port B
	movlw	b'00000000' ; All outputs ports low
	movwf	PORTA
	movwf	PORTB
; Wait and initialize HD44780
	call	delay_5		; Allow LCD time to initialize
						; itself
	call	delay_5
	call	initLCD		; Then do forced initialization
	call	delay_5		; Wait again
; Store base address of text buffer in PIC RAM

	movlw	0x0c		; Start address for buffer
	movwf	pic_ad		; to local variable
; Initialize EEPROM data to 0x0
	clrf	EEMemAdd	; Set address to 0
;======================
;   first LCD line
;======================
; Store 16 blanks in PIC RAM, starting at address stored
; in variable pic_ad
	call	blank16
; Call procedure to store ASCII characters for message
; in text buffer
	movlw	d'0'		; Offset into buffer
	call	storeMS1
;=======================
;  Read EEPROM memory
;=======================
; EEPROM memory address to use is at 10 (0x0a). Variable
; EEMemAdd is already initialized.
; Fill data for EEPROM is 0xff. This value indicates
; the first iteration  
	call 	EERead		; Local procedure. Value in w
	movwf	EEByte		; Save result
; EEPROM data still in w
	incf	EEByte,f
	call	EEWrite
; At this point iteration inNum is stored in EEByte
; This value must be displayed on the LCD at offset 11
; of the first line. This means it must be stored at offset
; 11 in the buffer. Since the buffer starts at 0x0c the
; iteration digit must be stored at offset 0x0c+11=0x17 
ShowEEData:
; Binary data in EEByte 
	movf	EEByte,w	; Value to w
	call	bin2asc		; Conversion routine
; At this point three ASCII digits are stored in local
; variables. Move digits to display area
	movf	asc1,w		; Unit digit
	movwf	0x18		; Store in buffer
	movf	asc10,w		; same with other digits
	movwf	0x17
	movf	asc100,w
	movwf	0x16
; Display line
; Set DDRAM address to start of first line
showLine:
	call 	line1
; Call procedure to display 16 characters in LCD
	call	display16 
loopHere:
	goto    loopHere  ;done

;============================================================
;                initialize LCD for 4-bit mode 
;============================================================
initLCD:
; Initialization for Densitron LCD module as follows:
;	4-bit interface
;   2 display lines of 16 characters each
;   cursor on
;   left-to-right increment
;   cursor shift right
;   no display shift
;=======================|
;   set command mode    |
;=======================|
	bcf		PORTA,E_line	; E line low
	bcf		PORTA,RS_line	; RS line low
	bcf		PORTA,RW_line	; Write mode
	call	delay_125		; delay 125 microseconds
;***********************|
;     FUNCTION SET      |
;***********************|
	movlw	0x28	; 0 0 1 0 1 0 0 0 (FUNCTION SET)
	call	send8	; 4-bit send routine

; Set 4-bit mode command must be repeated
	movlw	0x28
	call 	send8

;***********************|
; DISPLAY AND CURSOR ON |
;***********************|
	movlw	0x0e	; 0 0 0 0 1 1 1 0 (DISPLAY ON/OFF)
	call	send8
;***********************|
;   set entry mode      |
;***********************|
	movlw	0x06	; 0 0 0 0 0 1 1 0 (ENTRY MODE SET)
	call	send8

;***********************|
; cursor/display shift  |
;***********************|
	movlw	0x14	; 0 0 0 1 0 1 0 0 (CURSOR/DISPLAY SHIFT)
	call	send8
;***********************|
;    clear display      |
;***********************|
	movlw	0x01	; 0 0 0 0 0 0 0 1 (CLEAR DISPLAY)
	call	send8
; Per documentation
	call	delay_5	; Test for busy
	return

;=======================
;  Procedure to delay
;   42 microseconds
;=======================
delay_125
	movlw	D'42'			; Repeat 42 machine cycles
	movwf	count1			; Store value in counter
repeat
	decfsz	count1,f   		; Decrement counter
	goto	repeat			; Continue if not 0
	return					; End of delay	

;=======================
;  Procedure to delay
;   5 milliseconds
;=======================
delay_5
	movlw	D'41'			; Counter = 41
	movwf	count2			; Store in variable
delay
	call	delay_125		; Delay
	decfsz	count2,f    	; 40 times = 5 milliseconds
	goto	delay
	return				; End of delay
;========================
;     pulse E line 
;========================
pulseE
	bsf		PORTA,E_line	; Pulse E line
	nop
	bcf		PORTA,E_line
	return

;=============================
;   long delay sub-routine
;     (for debugging)
;=============================
long_delay
		movlw	D'200'	; w = 200 decimal
		movwf	J		; J = w
jloop:	movwf	K		; K = w
kloop:	decfsz	K,f		; K = K-1, skip next if zero
		goto 	kloop
		decfsz	J,f		; J = J-1, skip next if zero
		goto	jloop
		return
;=============================
;   LCD display procedure
;=============================
; Sends 16 characters from PIC buffer with address stored
; in variable pic_ad to LCD line previously selected
display16
	call	delay_5			; Make sure not busy
; Set up for data
	bcf		PORTA,E_line	; E line low
	bsf		PORTA,RS_line	; RS line high for data
; Set up counter for 16 characters
	movlw	D'16'			; Counter = 16
	movwf	count3	
; Get display address from local variable pic_ad
	movf	pic_ad,w		; First display RAM address to W
	movwf	FSR				; W to FSR
getchar
	movf	INDF,w			; get character from display RAM
							; location pointed to by file select
							; register
	call	send8			; 4-bit interface routine
; Test for 16 characters displayed
	decfsz	count3,f  		; Decrement counter
	goto	nextchar		; Skipped if done
	return
nextchar:
	incf	FSR,f			; Bump pointer
	goto	getchar

;========================
;   send 2 nibbles in
;     4-bit mode
;========================
; Procedure to send two 4-bit values to port B lines
; 7, 6, 5, and 4. High-order nibble is sent first
; ON ENTRY:
;         w register holds 8-bit value to send
send8:
	movwf	store1		; Save original value
	call	merge4		; Merge with port B
; Now w has merged byte
	movwf	PORTB		; w to port B
	call	pulseE		; Send data to LCD
; High nibble is sent
	movf	store1,w	; Recover byte into w
	swapf	store1,w	; Swap nibbles in w
	call	merge4
	movwf	PORTB
	call	pulseE		; Send data to LCD
	call	delay_125 
	return
;=================
;  merge bits
;=================
; Routine to merge the 4 high-order bits of the
; value to send with the contents of port B
; so as to preserve the 4 low-bits in port B
; Logic:
;      AND value with 1111 0000 mask
;      AND port B with 0000 1111 mask
;      Now low nibble in value and high nibble in
;      port B are al 0 bits:
;          value = vvvv 0000
;         port B = 0000 bbbb
;      OR value and port B resulting in:
;                  vvvv bbbb 
; ON ENTRY:
;      w contain value bits
; ON EXIT:
;      w contains merged bits
merge4:
	andlw	b'11110000'	; ANDing with 0 clears the
						; bit. ANDing with 1 preserves
						; the original value
	movwf	store2		; Save result in variable
	movf	PORTB,w		; port B to w register
	andlw	b'00001111' ; Clear high nibble in port b
						; and preserve low nibble
	iorwf	store2,w	; OR two operands in w
	return

;========================
;     blank buffer
;========================
; Procedure to store 16 blank characters in PIC RAM
; buffer starting at address stored in the variable
; pic_ad
blank16
	movlw	D'16'		; Setup counter
	movwf	count1
	movf	pic_ad,w	; First PIC RAM address
	movwf	FSR			; Indexed addressing
	movlw	0x20		; ASCII space character
storeit
	movwf	INDF		; Store blank character in PIC RAM
						; buffer using FSR register
	decfsz	count1,f  	; Done?
	goto	incfsr		; no
	return				; yes
incfsr
	incf	FSR,f		; Bump FSR to next buffer space
	goto	storeit

;========================
; Set address register
;    to LCD line 1
;========================
; ON ENTRY:
;         Address of LCD line 1 in constant LCD_1 
line1:
	bcf		PORTA,E_line 	; E line low
	bcf		PORTA,RS_line	; RS line low, set up for control
	call	delay_5		; busy?
; Set to second display line
	movlw	LCD_1			; Address and command bit
	call	send8			; 4-bit routine
; Set RS line for data
	bsf		PORTA,RS_line	; Setup for data
	call	delay_5			; Busy?
	return

;===============================
;  first text string procedure
;===============================
storeMS1:
; Procedure to store in PIC RAM buffer the message
; contained in the code area labeled msg1
; ON ENTRY:
;         variable pic_ad holds address of text buffer
;         in PIC RAM
;         w register hold offset into storage area
;         msg1 is routine that returns the string characters
;         an a zero terminator
;         index is local variable that hold offset into
;         text table. This variable is also used for
;         temporary storage of offset into buffer 
; ON EXIT:
;         Text message stored in buffer
;
; Store offset into text buffer (passed in the w register)
; in temporary variable 
	movwf	index		; Store w in index
; Store base address of text buffer in FSR
	movf	pic_ad,w	; first display RAM address to W
	addwf	index,w		; Add offset to address
	movwf	FSR			; W to FSR
; Initialize index for text string access
	movlw	0			; Start at 0
	movwf	index		; Store index in variable
; w still = 0
get_msg_char:
	call	msg1		; Get character from table
; Test for zero terminator
	andlw	0x0ff
	btfsc	STATUS,Z	; Test zero flag
	goto	endstr1		; End of string
; ASSERT: valid string character in w
;         store character in text buffer (by FSR)
	movwf	INDF		; store in buffer by FSR
	incf	FSR,f		; increment buffer pointer
; Restore table character counter from variable
	movf	index,w		; Get value into w
	addlw	1			; Bump to next character
	movwf	index		; Store table index in variable
	goto	get_msg_char	; Continue	
endstr1:
	return

; Routine for returning message stored in program area
; Message has 10 characters
msg1:
	addwf	PCL,f		; Access table
	retlw	'I'
	retlw	't'
	retlw	'e'
	retlw	'r'
	retlw	'.'
	retlw	0x20
	retlw	'N'
	retlw	'o'
	retlw	'.'
	retlw	0x20
	retlw	0

;==============================
;   binary to ASCII decimal
;        conversion
;============================== 
; ON ENTRY:
;         w register has binary value in range 0 to 255
; ON EXIT:
;         output variables asc100, asc10, and asc1 have
;         three ASCII decimal digits
; Routine logic:
;   The value 100 is subtracted from the source operand
;   until the reminder is < 0 (carry cleared). The number
;   of subtractions is the decimal hundreds result. 100 is
;   then added back to the subtrahend to compensate
;   for the last subtraction. Now 10 is subracted in the
;   same manner to determine the decimal tenths result.
;   The final reminder is the decimal units result.
; Variables:
;     inNum     storage for source operand
;     asc100    storage for hundreds position result
;     asc10     storage for tenth position result
;     asc1      storage for unit position reslt
;     thisDig	Digit counter 
bin2asc:
	movwf	inNum	; Save copy of source value
	clrf	asc100  ; Clear hundreds storage
	clrf	asc10	; Tens
	clrf	asc1	; Units
	clrf	thisDig
sub100:
	movlw	.100
	subwf	inNum,f   	; Subtract 100
	btfsc	STATUS,C   	; Did subtract overflow?
	goto	bump100		; No. Count subtraction
	goto	end100
bump100:
	incf	thisDig,f	;increment digit counter
	goto	sub100
; Store 100th digit
end100:
	movf	thisDig,w  	; Adjusted digit counter
	addlw	0x30		; Convert to ASCII
	movwf	asc100		; Store it
; Calculate tenth position value
	clrf	thisDig
; Adjust minuend
	movlw	.100		; Minuend
	addwf	inNum,f		; Add value to minuend to 
						; compensate for last operation
sub10:
	movlw	.10
	subwf	inNum,f   	; Subtract 10
	btfsc	STATUS,C   	; Did subtract overflow?
	goto	bump10		; No. Count subtraction
	goto	end10
bump10:
	incf	thisDig,f	;increment digit counter
	goto	sub10
; Store 10th digit
end10:
	movlw	.10
	addwf	inNum,f		; Adjust for last subratct
	movf	thisDig,w  	; get digit counter contents
	addlw	0x30		; Conver to ASCII
	movwf	asc10		; Store it
; Calculate and store units digit
	movf	inNum,w   ;	Store units value
	addlw	0x30		; Conver to ASCII
	movwf	asc1		; Store digit
	return

;============================================================
;                    EEPROM procedures
;============================================================
;==============================
;       read EEPROM
;==============================
; Procedure to read EEPROM memory. Address of memory
; location to read is stored in local register EEMemAdd
; On exit: read data in w
EERead:
	bcf 	STATUS,RP0 	; Bank 0
	movf 	EEMemAdd,w	; Address to w
	movwf	EEADR 		; w to address register
	bsf 	STATUS,RP0 	; Bank 1
	bsf 	EECON1,RD 	; EE Read
	bcf 	STATUS,RP0 	; Bank 0
	movf 	EEDATA,w 	; W = EEDATA
	return

;==============================
;         write EEPROM
;==============================
; Procedure to write asc1 byte to EEPROM memory
; Address to write stored in local register EEMemAdd
; Data byte to write is in local register EEByte
EEWrite:
; Load byte to write into EE data register
	movf	EEByte,w	; Data to w
	movwf	EEDATA		; Write	
; Set write address in EE address register
	movf 	EEMemAdd,w	; Address to w
	movwf	EEADR 		; w to address register
; Write data to EEPROM memory
	bsf 	STATUS,RP0 	; Bank 1
	bcf 	INTCON,GIE 	; Disable INTs.
	bsf 	EECON1,WREN ; Enable Write
	movlw 	0x55 		; Code # 1
	movwf 	EECON2 		; Write 0x55
	movlw 	0xaa 		; Code # 2
	movwf 	EECON2 		; Write 0xaa
	bsf 	EECON1,WR 	; Set WR bit
; Write operation now takes place automatically
	bsf 	INTCON,GIE 	; Re-enable interrupts
	bcf		STATUS,RP0	; Bank 0
	return

	end

