; File name: Key2LCD.asm
; Date: May 11, 2006
; Author: Julio Sanchez
;
; Description:
; Decode 4 x 4 keypad and display scan code in LCD.
; Program to uses 4-bit PIC-to-LCD interface.
; Code assumes that LCD is driven by Hitachi HD44780
; controller and PIC 16F977. Display supports two lines
; each one with 20 characters. The wiring and base
; address of each display line is stored in #define
; statements. These statements can be edited to
; accomodate a different set-up.
; Keypad switch wiring (values are scan codes):
;        --- KEYPAD -- 
;        0   1   2   3   <= port B0 |
;        4   5   6   7   <= port B1 |--- ROWS = OUTPUTS
;        8   9   A   B   <= port B2 |
;        C   D   E   F   <= port B3 |
;        |   |   |   |
;        |   |   |   |_____ port B4 |
;        |   |   |_________ port B5 |--- COLUMNS = INPUTS
;        |   |_____________ port B6 |
;        |_________________ port B7 |
;
; Program operations:
; 1. Key press action generates a scan code in the range
;    0x0 to 0xf.
; 2. Program convers scan code to ASCII digit and displays
;    the digit on the LCD.
; 3. When the end of the first LCD line is reached, display
;    continues in the second line. When the end of the
;    second line is reached, LCD is reset to the first line
;
; Program uses delay loops for interface timing.
; WARNING:
; Code assumes 4Mhz clock. Delay routines must be
; edited for faster clock
;
;===========================
;      16F877 switches
;===========================
; Switches used in __config directive:
;   _CP_ON     	  Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF  
;   _BODEN_ON     Brown-out reset enable ON/OFF
; * _BODEN_OFF 
; * _PWRTE_ON     Power-up timer enable ON/OFF
;   _PWRTE_OFF   
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF
;   _LPV_ON       Low voltage IC programming enable ON/OFF
; * _LPV_OFF
;   _CPD_ON       Data EE memory code protection ON/OFF
; * _CPD_OFF
; OSCILLATOR CONFIGURATIONS:        
;   _LP_OSC       Low power crystal occilator
;   _XT_OSC       External parallel resonator/crystal ocillator  
; * _HS_OSC       High speed crystal resonator
;   _RC_OSC       Resistor/capacitor ocillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

	processor 	16f877		; Define processor
	#include	<p16f877.inc>
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _HS_OSC & _WDT_OFF & _LVP_OFF & _CPD_OFF

; __CONFIG directive is used to embed configuration data
;  within the source file. The labels following the directive
;  are located in the corresponding .inc file.

;============================================================
;                   constant definitions
;    for PIC-to-LCD pin wiring and LCD line addresses
;============================================================
#define E_line 1   	;|
#define RS_line 0	;| -- from wiring diagram 
#define RW_line 2	;|
; LCD line addresses (from LCD data sheet)
#define LCD_1 0x80	; First LCD line constant
#define LCD_2 0xc0	; Second LCD line constant
#define LCDlimit .20; Number of characters per line
; Note: The constant that define the LCD display line
;       addresses have the high-order bit set in
;       order to faciliate the controller command 
;
;============================================================
;                    PIC register equates
;============================================================
portd	equ		0x08
porte   equ     0x09
fsr		equ		0x04
status  equ		0x03
indf	equ 	0x00
z		equ		2
c		equ		0
;=============================================================
;                  variables in PIC RAM
;=============================================================
; Reserve 20 bytes for string buffer
	cblock	0x20
	strData
	endc
; Leave 16 bytes and Continue with local variables
	cblock	0x34		; Start of block
	count1		; Counter # 1
	count2		; Counter # 2
	count3		; Counter # 3
	pic_ad		; Storage for start of text area
				; (labeled strData) in PIC RAM
	J			; counter J
	K			; counter K
	index		; Index into text table (also used
				; for auxiliary storage)
	store1		; Local storage
	store2
; For LCDscroll procedure
	LCDcount	; Counter for characters per line
	LCDline		; Current display line (0 or 1)
; Keypad processing variables
	keyMask		; For keypad processing
	rowMask		; For masking-off key rows
	rowCode		; Row addend for calculating scan code
	rowCount	; Counter for key rows (0 to 3)
	scanCode	; Final key code
	newScan		; 0 if no new scan code detected
	endc
;============================================================
;                          M A C R O S
;============================================================
; Macros to select the register banks
; Data memory bank selection bits:
; RP1:RP0          Bank
;   0:0             0    Ports A,B,C,D, and E
;   0:1             1    Tris A,B,C,D, and E
;   1:0             2
;   1:1             3
Bank0	MACRO			; Select RAM bank 0
		bcf	STATUS,RP0
		bcf	STATUS,RP1
		ENDM

Bank1	MACRO			; Select RAM bank 1
		bsf	STATUS,RP0
		bcf	STATUS,RP1
		ENDM

Bank2	MACRO			; Select RAM bank 2
		bcf	STATUS,RP0
		bsf	STATUS,RP1
		ENDM

Bank3	MACRO			; Select RAM bank 3
		bsf	STATUS,RP0
		bsf	STATUS,RP1
		ENDM
;============================================================
;                   M A I N    P R O G R A M
;============================================================
		org		0	  ; start at address 
		goto	main
; Space for interrupt handlers
	org		0x08
main:
; Wiring:
;     LCD data to port D, lines 0 to 7
;     E line -> port E, 1
;     RW line -> port E, 2
;     RS line -> port E, 0
; Set porte D and E for output
; First, initialize port B by clearing latches
	clrf	STATUS
	clrf	PORTB
; Select bank 1 to tris port D for output
	Bank1
; Tris port D for output. Port D lines 4 to 7 are wired
; to LCD data lines. Port D lines 0 to 4 are wired to LEDs.
	movlw	B'00000000'
	movwf	TRISD		; and port D
; By default port A lines are analog. To configure them
; as digital code must set bits 1 and 2 of the ADCON1
; register (in bank 1)
	movlw	0x06		; binary 0000 0110  is code to
						; make all port A lines digial
	movwf	ADCON1
; Port B, lines are wired to keypad swtiches, as follows:
;   7 6 5 4 3 2 1 0
;   | | | | |_|_|_|_____ switch rows (output)
;   |_|_|_|_____________ switch columns (input)
; rows must be defined as output and columns as input
	movlw	b'11110000'
	movwf	TRISB
; Tris port E for output
	movlw	B'00000000'
	movwf	TRISE		; Tris port E
; Enable port B pullups for switches in OPTION register
;   7  6  5  4  3  2  1  0 <= OPTION bits
;   |  |  |  |  |  |__|__|_____ PS2-PS0 (prescaler bits)
;   |  |  |  |  |               Values for Timer0
;   |  |  |  |  |               000 = 1:2   001 = 1:4
;   |  |  |  |  |               010 = 1:8   011 = 1:16
;   |  |  |  |  |               100 = 1:32  101 = 1:64
;   |  |  |  |  |               110 = 1:128 *111 = 1:256
;   |  |  |  |  |______________ PSA (prescaler assign)
;   |  |  |  |                 *1 = to WDT
;   |  |  |  |                  0 = to Timer0
;   |  |  |  |_________________ TOSE (Timer0 edge select)
;   |  |  |                    *0 = increment on low-to-high
;   |  |  |                     1 = increment in high-to-low
;   |  |  |____________________ TOCS (TMR0 clock source)
;   |  |                       *0 = internal clock
;   |  |                        1 = RA4/TOCKI bit source
;   |  |_______________________ INTEDG (Edge select)
;   |                          *0 = falling edge
;   |__________________________ RBPU (Pullup enable)
;                              *0 = enabled
;                               1 = dissabled
	movlw	b'00001000'
	movwf	OPTION_REG
; Back to bank 0
	Bank0
; Clear all output lines
	movlw	b'00000000' 
	movwf	portd
	movwf	porte
; Wait and initialize HD44780
	call	delay_5		; Allow LCD time to initialize itself
	call	initLCD		; Then do forced initialization
	call	delay_5		; (Wait probably not necessary)
; Set display address to start of second LCD line
	call	line1
; Clear character counter and line counter variables
	clrf	LCDcount
	clrf	LCDline
;========================
;      scan keypad  
;========================
; Keypad switch wiring:
;        x   x   x   x   <= port B0 |
;        x   x   x   x   <= port B1 |--- ROWS = OUTPUTS
;        x   x   x   x   <= port B2 |
;        x   x   x   x   <= port B3 |
;        |   |   |   |
;        |   |   |   |_____ port B4 |
;        |   |   |_________ port B5 |--- COLUMNS = INPUTS
;        |   |_____________ port B6 |
;        |_________________ port B7 |
; Swtiches are connected to port B lines
; Clear scan code register
	clrf	scanCode
;============================
;  scan keypad and display
;============================
keyScan:
; Port B, lines are wired to pushbutton swtiches, as follows:
;   7 6 5 4 3 2 1 0
;   | | | | |_|_|_|_____ switch rows (output)
;   |_|_|_|_____________ switch columns (input)
; Keypad processing:
; swtich rows are succesively grounded (row = 0)
; Then column values are tested. If a column returns 0
; in a 0 row, that switch is down.
; Initialize row code addend 
	clrf	rowCode			; First row is code 0
	clrf	newScan			; No new scan code detected 
; Initialize row count
	movlw	D'4'			; Four rows 
	movwf	rowCount		; Register variable
	movlw	b'11111110'		; All set but LOB
	movwf	rowMask
keyLoop:
; Initialize row eliminator mask:
; The row mask is ANDed with the key mask to succesively
; mask-off each row, for example:
;
;                  |------- row 3
;                  ||------ row 2
;                  |||----- row 1
;                  ||||---- row 0
;             0000 1111 <= key mask
;       AND   1111 1101 <= mask for row 1
;             ---------
;             0000 1101 <= row 1 is masked off
;
; The row mask, which is initally 1111 1110 is rotated left
; through the carry in order to mask off the next row
	movlw	b'00001111'		; Mask off all lines
	movwf	keyMask			; To local register 
; Set row mask for current row
	movf	rowMask,w		; Mask to w
	andwf	keyMask,f		; Update key mask
	movf	keyMask,w		; Key mask to w
	movwf	PORTB			; Mask-off port B lines
; Read port B lines 4 to 7 (columns are input)
	btfss	PORTB,4
	call	col0			; Key column procedures
	btfss	PORTB,5
	call	col1
	btfss	PORTB,6
	call	col2
	btfss	PORTB,7
	call	col3
; Index to next row by adding 4 to row code
	movf	rowCode,w		; Code to w
	addlw	D'4'
	movwf	rowCode
;=========================
;     shift row mask
;=========================
; Set the carry flag
	bsf		STATUS,c
	rlf		rowMask,f		; Rotate mask bits in storage
;=========================
;     end of keypad?
;=========================
; Test for last key row (maximum count is 4) 
	decfsz	rowCount,f  ; Decrement counter
	goto	keyLoop
;===========================================================
;===========================================================
;                   display scan code
;===========================================================
;===========================================================
; At this point all keys have been tested.
; Variable newScan = 0 if no new scan code detected, else
; variable scanCode holds scan code 
	movf	newScan,f		; Copy onto intsef (sets z flag)
	btfsc	STATUS,z		; Is it zero
	goto	scanExit
; At this point a new scan code is detected
	movf	scanCode,w		; To w 
; If scan code is in the range 0 to 9, that is, a decimal
; digit, then ASCII conversion consists of adding 0x30.
; If the scan code represents one of the hex letters
; (0xa to 0xf) then ASCII conversion requires adding
; 0x37
	sublw	0x09		; 9 - w
; if w from 0 to 9 then 9 - w = positive (c flag = 1)
; if w = 0xa then 9 - 10 = -1 (c flag = 0)
; if w = 0xc then 9 - 12 = -2 (c flag = 0)
	btfss	STATUS,c	; Test carry flag
	goto	hexLetter	; Carry clear, must be a letter 	
; At this point scan code is a decimal digit in the
; range 0 to 9. Conver to ASCII by adding 0x30
	movf	scanCode,w	; Recover scan code
	addlw	0x30		; Conver to ASCII
	goto	displayDig
hexLetter:
	movf	scanCode,w	; Recover scan code
	addlw	0x37		; Conver to ASCII
displayDig:
	call	send8		; Display routine
	call	LCDscroll	; Auto line scrolling procedure
scanExit:
	call	long_delay  ; Debounce
	goto    keyScan  	; Continue

;==========================
;   calculate scan code
;==========================
; The column position is added to the row code (stored
; in rowCode register). Sum is the scan code
col0:
	movf	rowCode,w	; Row code to w
	addlw	0x00		; Add 0 (clearly not necessary)
	movwf	scanCode	; Final value
	incf	newScan,f		; New scan code
	return

col1:
	movf	rowCode,w	; Row code to w
	addlw	0x01		; Add 1
	movwf	scanCode
	incf	newScan,f
	return

col2:
	movf	rowCode,w	; Row code to w
	addlw	0x02		; Add 2
	movwf	scanCode
	incf	newScan,f
	return

col3:
	movf	rowCode,w	; Row code to w
	addlw	0x03		; Add 3
	movwf	scanCode
	incf	newScan,f
	return

;============================================================
;                initialize LCD for 4-bit mode 
;============================================================
initLCD:
; Initialization for Densitron LCD module as follows:
;	4-bit interface
;   2 display lines of 16 characters each
;   cursor on
;   left-to-right increment
;   cursor shift right
;   no display shift
;=======================|
;   set command mode    |
;=======================|
	bcf		porte,E_line	; E line low
	bcf		porte,RS_line	; RS line low
	bcf		porte,RW_line	; Write mode
	call	delay_125		; delay 125 microseconds
;***********************|
;     FUNCTION SET      |
;***********************|
	movlw	0x28	; 0 0 1 0 1 0 0 0 (FUNCTION SET)
					;     | | | |__ font select:
					;     | | |    1 = 5x10 in 1/8 or 1/11 dc
					;     | | |    0 = 1/16 dc
					;     | | |___ Duty cycle select
					;     | |      0 = 1/8 or 1/11
					;     | |      1 = 1/16 (multiple lines)
					;     | |___ Interface width
					;     |      0 = 4 bits
					;     |      1 = 8 bits
					;     |___ FUNCTION SET COMMAND 
	call	send8	; 4-bit send routine

; Set 4-bit mode command must be repeated
	movlw	0x28
	call 	send8

;***********************|
; DISPLAY AND CURSOR ON |
;***********************|
	movlw	0x0e	; 0 0 0 0 1 1 1 0 (DISPLAY ON/OFF)
					;         | | | |___ Blink character at cursor
					;         | | |      1 = on, 0 = off
					;         | | |___ Curson on/off
					;         | |      1 = on, 0 = off
					;         | |____ Display on/off
					;         |       1 = on, 0 = off
					;         |____ COMMAND BIT
	call	send8
;***********************|
;   set entry mode      |
;***********************|
	movlw	0x06	; 0 0 0 0 0 1 1 0 (ENTRY MODE SET)
					;           | | |___ display shift
					;           | |      1 = shift
					;           | |      0 = no shift
					;           | |____ cursor increment mode
					;           |       1 = left-to-right
					;           |       0 = right-to-left
					;           |___ COMMAND BIT
	call	send8

;***********************|
; cursor/display shift  |
;***********************|
	movlw	0x14	; 0 0 0 1 0 1 0 0 (CURSOR/DISPLAY SHIFT)
   					;       | | | |_|___ don't care
					;       | |_|__ cursor/display shift
					;       |       00 = cursor shift left
					;       |       01 = cursor shift right
					;       |       10 = cursor and display
					;       |            shifted left
					;       |       11 = cursor and display
					;       |            shifted right
					;       |___ COMMAND BIT
	call	send8
;***********************|
;    clear display      |
;***********************|
	movlw	0x01	; 0 0 0 0 0 0 0 1 (CLEAR DISPLAY)
					;               |___ COMMAND BIT
	call	send8
; Per documentation
	call	delay_5	; Test for busy
	return

;=======================
;  Procedure to delay
;   42 microseconds
;=======================
delay_125:
	movlw	D'42'			; Repeat 42 machine cycles
	movwf	count1			; Store value in counter
repeat
	decfsz	count1,f   		; Decrement counter
	goto	repeat			; Continue if not 0
	return					; End of delay	

;=======================
;  Procedure to delay
;   5 milliseconds
;=======================
delay_5:
	movlw	D'42'			; Counter = 41
	movwf	count2			; Store in variable
delay
	call	delay_125		; Delay
	decfsz	count2,f    	; 40 times = 5 milliseconds
	goto	delay
	return				; End of delay
;========================
;     pulse E line 
;========================
pulseE
	bsf		porte,E_line	; Pulse E line
	nop
	bcf		porte,E_line
	return

;=============================
;   long delay sub-routine
;     (for debugging)
;=============================
long_delay
		movlw	D'200'	; w delay count
		movwf	J		; J = w
jloop:	movwf	K		; K = w
kloop:	decfsz	K,f		; K = K-1, skip next if zero
		goto 	kloop
		decfsz	J,f		; J = J-1, skip next if zero
		goto	jloop
		return

;========================
;   send 2 nibbles in
;     4-bit mode
;========================
; Procedure to send two 4-bit values to port B lines
; 7, 6, 5, and 4. High-order nibble is sent first
; ON ENTRY:
;         w register holds 8-bit value to send
send8:
	movwf	store1		; Save original value
	call	merge4		; Merge with port B
; Now w has merged byte
	movwf	portd		; w to port D
	call	pulseE		; Send data to LCD
; High nibble is sent
	movf	store1,w	; Recover byte into w
	swapf	store1,w	; Swap nibbles in w
	call	merge4
	movwf	portd
	call	pulseE		; Send data to LCD
	call	delay_125 
	return
;=================
;  merge bits
;=================
; Routine to merge the 4 high-order bits of the
; value to send with the contents of port B
; so as to preserve the 4 low-bits in port B
; Logic:
;      AND value with 1111 0000 mask
;      AND port B with 0000 1111 mask
;      Now low nibble in value and high nibble in
;      port B are al 0 bits:
;          value = vvvv 0000
;         port B = 0000 bbbb
;      OR value and port B resulting in:
;                  vvvv bbbb 
; ON ENTRY:
;      w contain value bits
; ON EXIT:
;      w contains merged bits
merge4:
	andlw	b'11110000'	; ANDing with 0 clears the
						; bit. ANDing with 1 preserves
						; the original value
	movwf	store2		; Save result in variable
	movf	portd,w		; port B to w register
	andlw	b'00001111' ; Clear high nibble in port b
						; and preserve low nibble
	iorwf	store2,w	; OR two operands in w
	return

;========================
; Set address register
;    to LCD line 1
;========================
; ON ENTRY:
;         Address of LCD line 2 in constant LCD_2 
line1:
	bcf		porte,E_line	; E line low
	bcf		porte,RS_line	; RS line low, setup for control
	call	delay_5		; Busy?
; Set to second display line
	movlw	LCD_1			; Address with high-bit set
	call	send8
; Set RS line for data
	bsf		porte,RS_line	; RS = 1 for data
	call	delay_5			; Busy?
	return

;========================
; Set address register
;    to LCD line 2
;========================
; ON ENTRY:
;         Address of LCD line 2 in constant LCD_2 
line2:
	bcf		porte,E_line	; E line low
	bcf		porte,RS_line	; RS line low, setup for control
	call	delay_5		; Busy?
; Set to second display line
	movlw	LCD_2			; Address with high-bit set
	call	send8
; Set RS line for data
	bsf		porte,RS_line	; RS = 1 for data
	call	delay_5			; Busy?
	return

;==========================
;   scroll to LCD line 2 
;==========================
; Procedure to count the number of characters displayed on
; each LCD line. If the number reaches the value in the
; constant LCDlimit, then display is scrolled to the second
; LCD line. If at the end of the second line, then LCD is
; reset to the first line.
LCDscroll:
	incf	LCDcount,f		; Bump counter
; Test for line limit
	movf	LCDcount,w
	sublw	LCDlimit		; Count minus limit
	btfss	STATUS,z		; Is count - limit = 0
	goto	scrollExit		; Go if not at end of line
; At this point the end of the LCD line was reached
; Test if this is also the end of the second line
	movf	LCDline,w
	sublw	0x01			; Is it line 1?
	btfsc	STATUS,z		; Is LCDline minus 1 = 0?
	goto	line2End		; Go if end of second line
; At this point it is the end of the top LCD line
	call	line2			; Scroll to second line
	clrf	LCDcount		; Reset counter
	incf	LCDline,f		; Bump line counter
	goto	scrollExit
; End of second LCD line
line2End:
	call	initLCD			; Reset
	clrf	LCDcount		; Clear counters
	clrf	LCDline
	call	line1			; Display to first line
scrollExit:
	return


	end

