; File: LapseTimer.ASM
; Date: May 1, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Using timer0 to produce a variable-lapse delay.
; The delay is calculated based on the number of machine
; cycles necessary for the desired wait period. For
; example, a machine running at a 4 Mhz clock rate
; executes 1,000,000 instructions per second. In this
; case a 1/2 second delay requires 500,000 instructions.
; The wait period is passed to the delay routine in three
; program registers whish hold the high-, middle-, and
; low-order bytes of the counter.
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator (simplest, 20% error)
; |
; |_____ * indicates setup values

	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF
;=====================================================
;              variables in PIC RAM
;=====================================================
; Local variables
	cblock	0x0d		; Start of block
		; 3-byte auxiliary counter for delay.
		countH      ; High-order byte
		countM		; Medium-order byte
		countL		; Low-order byte
	endc
;========================================================
;                 m a i n   p r o g r a m
;========================================================
	org	0	  ; start at address 0
	goto	main
;
;=============================
;      interrupt handler
;=============================
	org		0x04
;	goto	IntServ
;=============================
;       main program
;=============================
main:
; Clear the watchdog timer and reset prescaler
	clrf	TMR0
	clrwdt
; Set up the OPTION regiser bit map
	movlw	b'11010000'
;   7  6  5  4  3  2  1  0 <= OPTION bits
;   |  |  |  |  |  |__|__|_____ PS2-PS0 (prescaler bits)
;   |  |  |  |  |               Values for Timer0
;   |  |  |  |  |              *000 = 1:2   001 = 1:4
;   |  |  |  |  |               010 = 1:8   011 = 1:16
;   |  |  |  |  |               100 = 1:32  101 = 1:64
;   |  |  |  |  |               110 = 1:128 *111 = 1:256
;   |  |  |  |  |______________ PSA (prescaler assign)
;   |  |  |  |                  1 = to WDT
;   |  |  |  |                 *0 = to Timer0
;   |  |  |  |_________________ TOSE (Timer0 edge select)
;   |  |  |                     0 = increment on low-to-high
;   |  |  |                    *1 = increment in high-to-low
;   |  |  |____________________ TOCS (TMR0 clock source)
;   |  |                       *0 = internal clock
;   |  |                        1 = RA4/TOCKI bit source
;   |  |_______________________ INTEDG (Edge select)
;   |                          *0 = falling edge
;   |__________________________ RBPU (Pullup enable)
;                               0 = enabled
;                              *1 = dissabled
	option
; Setup ports
	movlw	0x00		; Set port B to output
	tris	PORTB
	clrf	PORTB		; All port B to 0
; Port A is not used in this program
;============================
;     display loop
;============================  
mloop:
; Turn on LED
	bsf		PORTB,0
; Initialize counters and delay
	call	onehalfSec
	call	TM0delay
; Turn off LED
	bcf		PORTB,0
; Re-initialize counter and delay
	call 	onehalfSec
	call	TM0delay
	goto	mloop
;==================================
;  variable-lapse delay procedure
;        using Timer0
;==================================
; ON ENTRY:
;         Variables countL, countM, and countH hold
;         the low-, middle-, and high-order bytes
;         of the delay period, in timer units 
; Routine logic:
; The prescaler is assigned to timer0 and setup so
; that the timer runs at 1:2 rate. This means that
; every time the counter reaches 128 (0x80) a total
; of 256 machine cycles have ellapsed. The value 0x80
; is detected by testing bit 7 of the counter
; register.
TM0delay:
; Note:
;     The timer0 register provides the low-order level
; of the count. Since the counter counts up from zero,
; in order to ensure that the initial low-level delay
; count is correct the value 128 - (xx/2) must be calculated
; where xx is the value in the original countL register.
; First calculate xx/2 by bit shifting
	bcf		STATUS,C	; Clear carry flag
	rrf		countL,f	; Divide by 2
; now subtract 128 - (xx/2)
	movf	countL,w	; w holds low-order byte
	sublw	d'128'
; Now w has adjusted result. Store in TMR0
	movwf	TMR0
; Routine tests timer overflow by testing bit 7 of
; the TMR0 register.
cycle:
	btfss	TMR0,7		; Is bit 7 set?
	goto	cycle		; Wait if not set
; At this point TMR0 bit 7 is set
; Clear the bit
	bcf		TMR0,7		; All other bits are preserved
; Subtract 256 from beat counter by decrementing the
; mid-order byte
	decfsz	countM,f
	goto	cycle		; Continue if mid-byte not zero
; At this point the mid-order byte has overflowed.
; High-order byte must be decremented.
	decfsz	countH,f	
	goto	cycle
; At this point the time cycle has ellapsed	
	return
;==============================
;  set register variables for
;     one-half second delay
;==============================
; Procedure to initialize local variables for a
; delay of one-half second on a 16F84 at 4 Mhz.
; Timer is setup for a 500,000 clock beats as
; follows: 500,000 = 0x07 0xa1 0x20
; 500,000 = 0x07 0xa1 0x20
;           ---- ---- ----
;            |    |    |___ countL)
;            |    |________ countM
;            |_____________ countH
onehalfSec:
	movlw	0x07
	movwf	countH
	movlw	0xa1
	movwf	countM
	movlw	0x20
	movwf	countL
	return

		end	

