; File: LapseTmrInt.ASM
; Date: May 1, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Interrupt-driven version of the LapseTimer program.
; Using timer0 to produce a variable-lapse delay.
; The delay is calculated based on the number of machine
; cycles necessary for the desired wait period. For
; example, a machine running at a 4 Mhz clock rate
; executes 1,000,000 instructions per second. In this
; case a 1/2 second delay requires 500,000 instructions.
; The wait period is passed to the delay routine in three
; register variables whish hold the high-, middle-, and
; low-order bytes of the counter.
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator (simplest, 20% error)
; |
; |_____ * indicates setup values

	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF

;=====================================================
;              variables in PIC RAM
;=====================================================
; Local variables
	cblock	0x0d		; Start of block
		; 3-byte auxiliary counter for delay.
		countH      ; High-order byte
		countM		; Medium-order byte
		countL		; Low-order byte
		old_w		; Context saving 
		old_STATUS  ; Idem
	endc
;========================================================
;                 m a i n   p r o g r a m
;========================================================
	org	0	  ; start at address 0
	goto	main
;
;=============================
;      interrupt handler
;=============================
	org		0x04
	goto	IntServ
;=============================
;       main program
;=============================
main:
; Clear the watchdog timer and reset prescaler
	clrf	TMR0
	clrwdt
; Set up the OPTION regiser bit map
	movlw	b'11011000'
;   7  6  5  4  3  2  1  0 <= OPTION bits
;   |  |  |  |  |  |__|__|_____ PS2-PS0 (prescaler bits)
;   |  |  |  |  |               Values for Timer0
;   |  |  |  |  |               000 = 1:2   001 = 1:4
;   |  |  |  |  |               010 = 1:8   011 = 1:16
;   |  |  |  |  |               100 = 1:32  101 = 1:64
;   |  |  |  |  |               110 = 1:128 *111 = 1:256
;   |  |  |  |  |______________ PSA (prescaler assign)
;   |  |  |  |                 *1 = to WDT
;   |  |  |  |                  0 = to Timer0
;   |  |  |  |_________________ TOSE (Timer0 edge select)
;   |  |  |                     0 = increment on low-to-high
;   |  |  |                    *1 = increment in high-to-low
;   |  |  |____________________ TOCS (TMR0 clock source)
;   |  |                       *0 = internal clock
;   |  |                        1 = RA4/TOCKI bit source
;   |  |_______________________ INTEDG (Edge select)
;   |                          *0 = falling edge
;   |__________________________ RBPU (Pullup enable)
;                               0 = enabled
;                              *1 = dissabled
	option
; Setup ports
	movlw	0x00		; Set port B to output
	tris	PORTB
	clrf	PORTB		; All port B to 0
; Port A is not used in this program
;============================
;      setup interrupts
;============================
; Clear external interrupt flag (INTF = bit 1)
	bcf		INTCON,INTF	; Clear flag
; Enable global interrupts (GIE = bit 7)
; Enable RB0 interrupt (inte = bit 4) 
	bsf		INTCON,GIE	; Enable global int (bit 7)
	bsf		INTCON,T0IE	; Enable TMR0 overflow interrupt
; Init count
	call	onehalfSec
;============================
;     do-nothing loop
;============================  
; All work is performed by the interrupt handler
mloop:
	goto	mloop
;==============================
;  set register variables for
;     one-half second delay
;==============================
; Procedure to initialize local variables for a
; delay of one-half second on a 16F84 at 4 Mhz.
; Timer is setup for a 500,000 clock beats as
; follows: 500,000 = 0x07 0xa1 0x20
; 500,000 = 0x07 0xa1 0x20
;           ---- ---- ----
;            |    |    |___ countL)
;            |    |________ countM
;            |_____________ countH
onehalfSec:
	movlw	0x07
	movwf	countH
	movlw	0xa1
	movwf	countM
	movlw	0x20
	movwf	countL
; The timer0 register provides the low-order level
; of the count. Since the counter counts up from zero,
; in order to ensure that the initial low-level delay
; count is correct the value 256 - xx must be calculated
; where xx is the value in the original countL register.
	movf	countL,w	; w holds low-order byte
	sublw	d'255'
; Now w has adjusted result. Store in TMR0
	movwf	TMR0
	return
;=======================================================
;           Interrupt Service Routine
;=======================================================
; Service routine receives control when there the timer
; register TMR0 overflows, that is, when 256 timer beats
; have ellapsed
IntServ:
; First test if source is a timer0 interrupt
	btfss	INTCON,T0IF	; T0IF is timer0 interrupt
	goto	notTOIF     ; Go if not RB0 origin 
; If so clear the timer interrupt flag so that count continues
	bcf		INTCON,T0IF	; Clear interrupt flag
; Save context
	movwf	old_w		; Save w register
	swapf	STATUS,w	; STATUS to w
	movwf	old_STATUS	; Save STATUS
;=========================
;   interrupt action
;=========================
; Subtract 256 from beat counter by decrementing the
; mid-order byte
	decfsz	countM,f
	goto	exitISR		; Continue if mid-byte not zero
; At this point the mid-order byte has overflowed.
; High-order byte must be decremented.
	decfsz	countH,f	
	goto	exitISR
; At this point count has expired so the programmed time
; has ellapsed. Service routine turns the LED on line 0,
; port B on and off at every conclusion of the count.
; This is done by xoring a mask with a one-bit at the
; port B line 0 position
	movlw	b'00000001'		; Xoring with a 1-bit produces
							; the complement
	xorwf	PORTB,f			; Complement bit 2, port B 
; Reset one-half second counter
	call	onehalfSec
;=========================
;        exit ISR
;=========================
exitISR:
; Restore context
	swapf		old_STATUS,w  ; Saved STATUS to w
	movfw		STATUS	; To STATUS register
	swapf		old_w,f	; Swap file register in itself
	swapf		old_w,w	; re-swap back to w
; Reset,interrupt
notTOIF:
	retfie	


		end	

