; File name: RTC2LCD.asm
; Last Update: June 6, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Program to demonstrate use of the NJU6355 Real Time Clock
; IC. Program uses LCD to display results of hours, minutes,
; and seconds, as follows:
;
; Top LCD line:   H:xx M:yy S:zz
;
; Initialization values are in #define statements that start
; with i, such as iYear, iMonth, etc.
;
; For LCD display parameters see the LCDTest2 program. 
; WARNING:
; Code assumes 4Mhz clock. Delay routines must be
; edited for faster clock
;
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

;=========================
; setup and configuration
;=========================
	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF

	errorlevel -302
; Supress bank-related warning
;============================================================
;                          M A C R O S
;============================================================
; Macros to select the register banks in 16F84
Bank0	MACRO			; Select RAM bank 0
		bcf		STATUS,RP0
		ENDM

Bank1	MACRO			; Select RAM bank 1
		bsf		STATUS,RP0
		ENDM 
;=====================================================
;                 constant definitions
;  for PIC-to-LCD pin wiring and LCD line addresses
;=====================================================
#define E_line 1   	;|
#define RS_line 2	;| -- from circuit wiring diagram 
#define RW_line 3	;|
; LCD line addresses (from LCD data sheet)
#define LCD_1 0x80	; First LCD line constant
#define LCD_2 0xc0	; Second LCD line constant
; Note: The constant that define the LCD display line
;       addresses have the high-order bit set in
;       order to faciliate the controller command 

; Defines from real-time clock wiring diagram
; all lines in port B
#define	DAT		0   ;|
#define	CLK		1   ;| -- from circuit wiring diagram
#define	CE		2   ;|
#define	IO		3   ;|
;
; Defines for RTC initialization (values are arbitrary)
#define iYear    .7
#define iMonth   .6	
#define iDay     .5
#define iDoW     .4
#define iHour    .3
#define iMin     .2
#define iSec     .1
;=====================================================
;                  PIC register equates
;=====================================================
;=====================================================
;              variables in PIC RAM
;=====================================================
; Reserve 16 bytes for string buffer
	cblock	0x0c
	strData
	endc
; Reserve three bytes for ASCII digits
	cblock	0x1d
	asc100
	asc10
	asc1 
	endc
; Continue with local variables
	cblock	0x20		; Start of block
	count1		; Counter # 1
	count2		; Counter # 2
	count3		; Counter # 3
	pic_ad		; Storage for start of text area
	J			; counter J
	K			; counter K
	index		; Index into text table (also used
				; for auxiliary storage)
	store1		; Local temporary storage 
	store2		; Storage # 2
; Storage for BCD digits
	bcdLow		; low-order nibble of packed BCD
	bcdHigh		; High-order nibble
; Variables for Real-Time Clock
	year
	month
	day
	dayOfWeek	; Sunday to Saturday (1 to 7)
	hour
	minutes
	seconds
	temp1
	counter
; Storage for BCD conversion routine
	inNum		; Source operand
	thisDig		; Digit counter
	endc

;=========================================================
;                           program
;=========================================================
		org		0	  ; start at address 
		goto	main
; Space for interrupt handlers
	org		0x08

main:
	movlw	b'00000000' ; All lines to output
	Bank1
	movwf	TRISA		; in port A
	movwf	TRISB		; and port B
	Bank0
	movlw	b'00000000' ; All outputs ports low
	movwf	PORTA
	movwf	PORTB
; Wait and initialize HD44780
	call	delay_5		; Allow LCD time to initialize
						; itself
	call	delay_5
	call	initLCD		; Then do forced initialization
	call	delay_5		; Wait again
; Store base address of text buffer in PIC RAM
	movlw	0x0c		; Start address for buffer
	movwf	pic_ad		; to local variable
;======================
;   first LCD line
;======================
; Store 16 blanks in PIC RAM, starting at address stored
; in variable pic_ad
	call	blank16
; Call procedure to store ASCII characters for message
; in text buffer
	movlw	d'0'		; Offset into buffer
	call	storeMS1	; Store message text in buffer
; Initialize real time clock
	call	initRTC		; Initialize variables
	call	setRTC		; Start clock
	call	delay_5		; Wait for operation to conclude
newTime:
; Get variables from RTC
	call	Get_Time
	call	delay_5		; Wait
	movf	hour,w		; Get hours
	call	Bcd2asc		; Conversion routine
; At this point three ASCII digits are stored in local
; variables. Move digits to display area
	movf	asc1,w		; Unit digit
	movwf	.15			; Store in buffer
	movf	asc10,w		; Same with other digit
	movwf	.14
	call	delay_5

	movf	minutes,w
	call	Bcd2asc		; Conversion routine
; At this point three ASCII digits are stored in local
; variables. Move two digits to display area
	movf	asc1,w		; Unit digit
	movwf	.20			; Store in buffer
	movf	asc10,w		; same with other digit
	movwf	.19
	call	delay_5

	movf	seconds,w
	call	Bcd2asc		; Conversion routine
; Move digits to display area
	movf	asc1,w		; Unit digit
	movwf	.25			; Store in buffer
	movf	asc10,w		; same with other digit
	movwf	.24
	call	delay_5
; Set DDRAM address to start of first line
	call 	line1
; Call procedure to display 16 characters in LCD
	call	display16 
	goto	newTime

;============================================================
;                initialize LCD for 4-bit mode 
;============================================================
initLCD:
; Initialization for Densitron LCD module as follows:
;	4-bit interface
;   2 display lines of 16 characters each
;   cursor on
;   left-to-right increment
;   cursor shift right
;   no display shift
;=======================|
;   set command mode    |
;=======================|
	bcf		PORTA,E_line	; E line low
	bcf		PORTA,RS_line	; RS line low
	bcf		PORTA,RW_line	; Write mode
	call	delay_125		; delay 125 microseconds
;***********************|
;     FUNCTION SET      |
;***********************|
	movlw	0x28	; 0 0 1 0 1 0 0 0 (FUNCTION SET)
	call	send8	; 4-bit send routine

; Set 4-bit mode command must be repeated
	movlw	0x28
	call 	send8

;***********************|
; DISPLAY AND CURSOR ON |
;***********************|
	movlw	0x0e	; 0 0 0 0 1 1 1 0 (DISPLAY ON/OFF)
	call	send8
;***********************|
;   set entry mode      |
;***********************|
	movlw	0x06	; 0 0 0 0 0 1 1 0 (ENTRY MODE SET)
	call	send8

;***********************|
; cursor/display shift  |
;***********************|
	movlw	0x14	; 0 0 0 1 0 1 0 0 (CURSOR/DISPLAY SHIFT)
	call	send8
;***********************|
;    clear display      |
;***********************|
	movlw	0x01	; 0 0 0 0 0 0 0 1 (CLEAR DISPLAY)
	call	send8
; Per documentation
	call	delay_5	; Test for busy
	return

;=======================
;  Procedure to delay
;   42 microseconds
;=======================
delay_125
	movlw	D'42'			; Repeat 42 machine cycles
	movwf	count1			; Store value in counter
repeat
	decfsz	count1,f   		; Decrement counter
	goto	repeat			; Continue if not 0
	return					; End of delay	

;=======================
;  Procedure to delay
;   5 milliseconds
;=======================
delay_5
	movlw	D'41'			; Counter = 41
	movwf	count2			; Store in variable
delay
	call	delay_125		; Delay
	decfsz	count2,f    	; 40 times = 5 milliseconds
	goto	delay
	return				; End of delay
;========================
;     pulse E line 
;========================
pulseE
	bsf		PORTA,E_line	; Pulse E line
	nop
	bcf		PORTA,E_line
	return

;=============================
;   long delay sub-routine
;     (for debugging)
;=============================
long_delay
		movlw	D'200'	; w = 200 decimal
		movwf	J		; J = w
jloop:	movwf	K		; K = w
kloop:	decfsz	K,f		; K = K-1, skip next if zero
		goto 	kloop
		decfsz	J,f		; J = J-1, skip next if zero
		goto	jloop
		return
;=============================
;   LCD display procedure
;=============================
; Sends 16 characters from PIC buffer with address stored
; in variable pic_ad to LCD line previously selected
display16
	call	delay_5			; Make sure not busy
; Set up for data
	bcf		PORTA,E_line	; E line low
	bsf		PORTA,RS_line	; RS line high for data
; Set up counter for 16 characters
	movlw	D'16'			; Counter = 16
	movwf	count3	
; Get display address from local variable pic_ad
	movf	pic_ad,w		; First display RAM address to W
	movwf	FSR				; W to FSR
getchar
	movf	INDF,w			; get character from display RAM
							; location pointed to by file select
							; register
	call	send8			; 4-bit interface routine
; Test for 16 characters displayed
	decfsz	count3,f  		; Decrement counter
	goto	nextchar		; Skipped if done
	return
nextchar:
	incf	FSR,f			; Bump pointer
	goto	getchar

;========================
;   send 2 nibbles in
;     4-bit mode
;========================
; Procedure to send two 4-bit values to port B lines
; 7, 6, 5, and 4. High-order nibble is sent first
; ON ENTRY:
;         w register holds 8-bit value to send
send8:
	movwf	store1		; Save original value
	call	merge4		; Merge with port B
; Now w has merged byte
	movwf	PORTB		; w to port B
	call	pulseE		; Send data to LCD
; High nibble is sent
	movf	store1,w	; Recover byte into w
	swapf	store1,w	; Swap nibbles in w
	call	merge4
	movwf	PORTB
	call	pulseE		; Send data to LCD
	call	delay_125 
	return
;=================
;  merge bits
;=================
; Routine to merge the 4 high-order bits of the
; value to send with the contents of port B
; so as to preserve the 4 low-bits in port B
; Logic:
;      AND value with 1111 0000 mask
;      AND port B with 0000 1111 mask
;      Now low nibble in value and high nibble in
;      port B are al 0 bits:
;          value = vvvv 0000
;         port B = 0000 bbbb
;      OR value and port B resulting in:
;                  vvvv bbbb 
; ON ENTRY:
;      w contain value bits
; ON EXIT:
;      w contains merged bits
merge4:
	andlw	b'11110000'	; ANDing with 0 clears the
						; bit. ANDing with 1 preserves
						; the original value
	movwf	store2		; Save result in variable
	movf	PORTB,w		; port B to w register
	andlw	b'00001111' ; Clear high nibble in port b
						; and preserve low nibble
	iorwf	store2,w	; OR two operands in w
	return

;========================
;     blank buffer
;========================
; Procedure to store 16 blank characters in PIC RAM
; buffer starting at address stored in the variable
; pic_ad
blank16
	movlw	D'16'		; Setup counter
	movwf	count1
	movf	pic_ad,w	; First PIC RAM address
	movwf	FSR			; Indexed addressing
	movlw	0x20		; ASCII space character
storeit
	movwf	INDF		; Store blank character in PIC RAM
						; buffer using FSR register
	decfsz	count1,f  	; Done?
	goto	incfsr		; no
	return				; yes
incfsr
	incf	FSR,f		; Bump FSR to next buffer space
	goto	storeit

;========================
; Set address register
;    to LCD line 1
;========================
; ON ENTRY:
;         Address of LCD line 1 in constant LCD_1 
line1:
	bcf		PORTA,E_line 	; E line low
	bcf		PORTA,RS_line	; RS line low, set up for control
	call	delay_5		; busy?
; Set to second display line
	movlw	LCD_1			; Address and command bit
	call	send8			; 4-bit routine
; Set RS line for data
	bsf		PORTA,RS_line	; Setup for data
	call	delay_5			; Busy?
	return

;===============================
;  first text string procedure
;===============================
storeMS1:
; Procedure to store in PIC RAM buffer the message
; contained in the code area labeled msg1
; ON ENTRY:
;         variable pic_ad holds address of text buffer
;         in PIC RAM
;         w register hold offset into storage area
;         msg1 is routine that returns the string characters
;         an a zero terminator
;         index is local variable that hold offset into
;         text table. This variable is also used for
;         temporary storage of offset into buffer 
; ON EXIT:
;         Text message stored in buffer
;
; Store offset into text buffer (passed in the w register)
; in temporary variable 
	movwf	index		; Store w in index
; Store base address of text buffer in FSR
	movf	pic_ad,w	; first display RAM address to W
	addwf	index,w		; Add offset to address
	movwf	FSR			; W to FSR
; Initialize index for text string access
	movlw	0			; Start at 0
	movwf	index		; Store index in variable
; w still = 0
get_msg_char:
	call	msg1		; Get character from table
; Test for zero terminator
	andlw	0x0ff
	btfsc	STATUS,Z	; Test zero flag
	goto	endstr1		; End of string
; ASSERT: valid string character in w
;         store character in text buffer (by FSR)
	movwf	INDF		; store in buffer by FSR
	incf	FSR,f		; increment buffer pointer
; Restore table character counter from variable
	movf	index,w		; Get value into w
	addlw	1			; Bump to next character
	movwf	index		; Store table index in variable
	goto	get_msg_char	; Continue	
endstr1:
	return

; Routine for returning message stored in program area
; Message has 10 characters
msg1:
	addwf	PCL,f		; Access table
	retlw	'H'
	retlw	':'
	retlw	' '
	retlw	' '
	retlw	' '
	retlw	'M'
	retlw	':'
	retlw	' '
	retlw	' '
	retlw   ' '
	retlw	'S'
	retlw	':'
	retlw	0

;==============================
;    BCD to ASCII decimal
;         conversion
;============================== 
; ON ENTRY:
;         w register has two packed BCD digits
; ON EXIT:
;         output variables asc10, and asc1 have
;         two ASCII decimal digits
; Routine logic:
;   The low order nibble is isolated and the value 30H
;   added to convert to ASCII. The result is stored in
;   the variable asc1. Then the same is done to the 
;   high-order nibble and the result is stored in the
;   variable asc10

Bcd2asc:
	movwf	store1		; Save input
	andlw	b'00001111' ; Clear high nibble
	addlw	0x30		; Convert to ASCII
	movwf	asc1		; Store result
	swapf	store1,w	; Recover input and swap digits
	andlw	b'00001111' ; Clear high nibble
	addlw	0x30		; Convert to ASCII
	movwf	asc10		; Store result
	return		

;============================================================
;                    6355 RTC procedures
;============================================================
;============================
;         init RTC
;============================
; Procedure to initialize the real time clock chip. If chip
; is not initialized it will not operate and the values
; read will be invalid.  
; Since the 6355 operates in BCD format the stored values must
; be converted to packed BCD.
; According to wiring diagram
; NJU6355 Interface for setting time:
; DAT		PORTB,0		Output
; CLK		PORTB,1		Output
; CE		PORTB,2		Output 
; IO		PORTB,3		Output
setRTC:
	Bank1
	movlw	b'00000000'	; All output bits
	movlw	TRISB
	Bank0
; Writing to the 6355 requires that the CLK bit be held
; low while the IO and CE lines are high
	bcf		PORTB,CLK	; CLK low
	call	delay_5	
	bsf		PORTB,IO	; IO high
	call	delay_5
	bsf		PORTB,CE	; CE high
; Data is stored in RTC as follows:
;  year    		8 bits (0 to 99)
;  month   		8 bits (1 to 12)
;  day    		8 bits (1 to 31)
;  dayOfWeek    4 bits (1 to 7)
;  hour   		8 bits (0 to 23)
;  minutes 		8 bits (0 to 59)
;        		======
;   Total 		44 bits
; Seconds cannot be written to RTC. RTC seconds register
; is automatically initialized to zero
	movf	year,w		; Get item from storage
	call	bin2bcd		; Convert to BCD
	movwf	temp1
	call	writeRTC

	movf	month,w
	call	bin2bcd
	movwf	temp1
	call	writeRTC

	movf	day,w
	call	bin2bcd
	movwf	temp1
	call	writeRTC

	movf	dayOfWeek,w		; dayOfWeek of week is 4-bits
	call	bin2bcd
	movwf	temp1
	call	write4RTC

	movf	hour,w
	call	bin2bcd
	movwf	temp1
	call	writeRTC

	movf	minutes,w
	call	bin2bcd
	movwf	temp1
	call	writeRTC
; Done
	bcf		PORTB,CLK		; Hold CLK line low
	call	delay_5
	bcf		PORTB,CE		; and the CE line                          ; to the RTC
	call	delay_5
	bcf		PORTB,IO		; RTC in output mode
	return
;============================
;      read RTC data
;============================
; Procedure to read the current time from the RTC and store
; data (in packed BCD format) in local time registers.
; According to wiring diagram
; NJU6355 Interface for read operations:
; DAT		PORTB,0		Input
; CLK		PORTB,1		Output
; CE		PORTB,2		Output
; IO		PORTB,3		Output
Get_Time
; Clear port B
	movlw	b'00000000'
	movwf	PORTB
; Make data line input
	Bank1
	movlw	b'00000001'
	movwf	TRISB
	Bank0
; Reading RTC data requires that the IO line be low and the
; CE line be high. CLK line is held low
	bcf		PORTB,CLK	; CLK low
	call	delay_125		
	bcf		PORTB,IO	; IO line low
	call	delay_125
	bsf		PORTB,CE	; and CE line high
; Data is read from RTC as follows:
;  year    		8 bits (0 to 99)
;  month   		8 bits (1 to 12)
;  day    		8 bits (1 to 31)
;  dayOfWeek    4 bits (1 to 7)
;  hour   		8 bits (0 to 23)
;  minutes 		8 bits (0 to 59)
;  seconds 		8 bits (0 to 59)
;        		======
;   Total 		52 bits
	call	readRTC
	movwf	year
	call	delay_125

	call	readRTC
	movwf	month
	call	delay_125

	call	readRTC
	movwf	day
	call	delay_125

; dayOfWeek of week is a 4-bit value
	call	read4RTC
	movwf	dayOfWeek
	call	delay_125

	call	readRTC
	movwf	hour
	call	delay_125

	call	readRTC
	movwf	minutes
	call	delay_125

	call	readRTC
	movwf	seconds

	bcf		PORTB,CE		; CE line low to end output
	return

;============================
;  read 4/8 bits from RTC 
;============================
; Procedure to read 4/8 bits stored in 6355 registers
; Value returned in w register
read4RTC
	movlw	.4		; 4 bit read
	goto	anyBits
readRTC
	movlw	.8		; 8 bits read
anyBits:
	movwf	counter
; Read 6355 read operation requires the IO line be set low
; and the CE line high. Data is read in the following order:
; year, month, day, day-of-week, hour, minutes, seconds
readBits:
	bsf		PORTB,CLK		; Set CLK high to validate data
	bsf		STATUS,C		; Set the carry flag (bit = 1)
; Operation:
;   If data line is high, then bit read is a 1-bit
;   otherwise bit read is a 0-bit
	btfss	PORTB,DAT		; Is data line high?
							; Leave carry set (1 bit) if high
	bcf		STATUS,C		; Clear the carry bit (make bit 0)
; At this point the carry bit matches the data line
	bcf		PORTB,CLK		; Set CLK low to end read
; The carry bit is now rotated into the temp1 register
	rrf		temp1,1
	decfsz	counter,1		; Decrement the bit counter 
	goto	readBits		; Continue if not last bit 
; At this point all bits have been read (8 or 4)
	movf	temp1,0			; Result to w
	return

;============================
;   write 4/8 bits to RTC
;============================
; Procedure to write 4 or 8 bits to the RTC registers
; ON ENTRY:
;    temp1 register holds value to be written
; ON EXIT:
;    nothing
write4RTC
	movlw	.4			; Init for 4 bits
	goto	allBits
writeRTC
	movlw	.8			; Init for 8 bits
allBits:
	movwf	counter		; Store in bit counter
writeBits:
	bcf		PORTB,CLK	; Clear the CLK line
	call	delay_5     ; Wait  
	bsf		PORTB,DAT	; Set the data line to RTC
	btfss	temp1,0		; Send LSB
	bcf		PORTB,DAT	; Clear data line
	call	delay_5		; Wait for operation to complete
	bsf		PORTB,CLK	; Bring CLK line high to validate
	rrf		temp1,f		; Rotate bits in storage 
	decfsz	counter,1	; Decrement bit counter
	goto	writeBits	; Continue if not last bit
	return

;============================
;   init time variables
;============================
; Procedure to initialize time variables for testing
; Constants used in ininitialization are located in
; #define statements.
initRTC:
	movlw	iYear
	movwf	year
	movlw	iMonth
	movwf	month
	movlw	iDay
	movwf	day
	movlw	iDoW
	movwf	dayOfWeek
	movlw	iHour
	movwf	hour
	movlw	iMin
	movwf	minutes
	movlw	iSec
	movwf	seconds
	return
;============================
;  binary to BCD conversion
;============================
; Convert a binary number into two packed BCD digits
; ON ENTRY:
;         w register has binary value in range 0 to 99
; ON EXIT:
;         output variables bcdLow and bcdHigh contain two
;         packed unpacked BCD digits
;         w contains two packed BCD digits
; Routine logic:
;   The value 10 is subtracted from the source operand
;   until the reminder is < 0 (carry cleared). The number
;   of subtractions is the high-order BCD digit. 10 is
;   then added back to the subtrahend to compensate
;   for the last subtraction. The final reminder is the
;   low-order BCD digit
; Variables:
;     inNum     storage for source operand
;     bcdHigh   storage for high-order nibble
;     bcdLow    storage for low-order nibble
;     thisDig	Digit counter 
bin2bcd:
	movwf	inNum	; Save copy of source value
	clrf	bcdHigh ; Clear storage
	clrf	bcdLow
	clrf	thisDig
min10:
	movlw	.10
	subwf	inNum,f   	; Subtract 10
	btfsc	STATUS,C   	; Did subtract overflow?
	goto	sum10		; No. Count subtraction
	goto	fin10
sum10:
	incf	thisDig,f	;increment digit counter
	goto	min10
; Store 10th digit
fin10:
	movlw	.10
	addwf	inNum,f		; Adjust for last subratct
	movf	thisDig,w  	; get digit counter contents
	movwf	bcdHigh		; Store it
; Calculate and store low-order BCD digit
	movf	inNum,w   	;	Store units value
	movwf	bcdLow		; Store digit
; Combine both digits
	swapf	bcdHigh,w	; High nibble to HOBs
	iorwf	bcdLow,w	; ORin low nibble
	return


	end

