; File name: RTC6355.asm
; Last revision: June 4, 2006
; Author: Julio Sanchez
; PIC: 16F877
;
; Description:
; Program to demonstrate programming of the NJU6355ED
; Real Time Clock IC. 
;
; Operation:
; 
; WARNING:
; Code assumes 10Mhz clock. Delay routines must be
; edited for faster clock. Clock speed also determines
; values for baud rate setting (see spbrgVal constant).
;
;===========================
;      16F877 switches
;===========================
; Switches used in __config directive:
;   _CP_ON     	  Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF  
;   _BODEN_ON     Brown-out reset enable ON/OFF
; * _BODEN_OFF 
; * _PWRTE_ON     Power-up timer enable ON/OFF
;   _PWRTE_OFF   
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF
;   _LPV_ON       Low voltage IC programming enable ON/OFF
; * _LPV_OFF
;   _CPD_ON       Data EE memory code protection ON/OFF
; * _CPD_OFF
; OSCILLATOR CONFIGURATIONS:        
;   _LP_OSC       Low power crystal occilator
;   _XT_OSC       External parallel resonator/crystal ocillator  
; * _HS_OSC       High speed crystal resonator
;   _RC_OSC       Resistor/capacitor ocillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

	processor 	16f877		; Define processor
	#include	<p16f877.inc>
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _HS_OSC & _WDT_OFF & _LVP_OFF & _CPD_OFF
; __CONFIG directive is used to embed configuration data
; within the source file. The labels following the directive
; are located in the corresponding .inc file.

	errorlevel -302
; Supress bank-related warning 
;============================================================
;                          M A C R O S
;============================================================
; Macros to select the register banks
Bank0	MACRO			; Select RAM bank 0
		bcf	STATUS,RP0
		bcf	STATUS,RP1
		ENDM

Bank1	MACRO			; Select RAM bank 1
		bsf	STATUS,RP0
		bcf	STATUS,RP1
		ENDM

Bank2	MACRO			; Select RAM bank 2
		bcf	STATUS,RP0
		bsf	STATUS,RP1
		ENDM

Bank3	MACRO			; Select RAM bank 3
		bsf	STATUS,RP0
		bsf	STATUS,RP1
		ENDM
;=====================================================
;                 constant definitions
;  for PIC-to-LCD pin wiring and LCD line addresses
;=====================================================
#define E_line 1   	;|
#define RS_line 0	;| -- from wiring diagram 
#define RW_line 2	;|
; LCD line addresses (from LCD data sheet)
#define LCD_1 0x80	; First LCD line constant
#define LCD_2 0xc0	; Second LCD line constant
#define LCDlimit .20; Number of characters per line
#define	spbrgVal .64; For 2400 baud on 10Mhz clock

; Defines from real-time clock wiring diagram
#define	CLK		PORTC,1
#define	DAT		PORTC,3
#define	IO		PORTC,5		; input/output select
#define	CE		PORTA,2		; chip enable bit

;=====================================================
;              variables in PIC RAM
;=====================================================
; Local variables
; Reserve 20 bytes for string buffer
	cblock	0x20
	strData
	endc
; Reserve three bytes for ASCII digits
	cblock	0x34
	asc100
	asc10
	asc1 
	endc
; Data
	cblock	0x37		; Start of block
	count1		; Counter # 1
	count2		; Counter # 2
	count3		; Counter # 3
	pic_ad
	J			; counter J
	K			; counter K
	index
	store1		; Local storage
	store2
; For LCDscroll procedure
	LCDcount	; Counter for characters per line
	LCDline		; Current display line (0 or 1)
; Communications variables
	newData		; not 0 if new data received
	ascVal	
	errorFlags
; Variables for Real-Time Clock
	Year
	Month
	Date
	Day
	Hours
	Minutes
	Seconds
	Time_Temp
	Time_Count
	endc

; EEPROM-related variables are placed in common area so
; they may be accessed from any bank
	cblock	0x70	
	EEMemAdd	; EEPROM address to access
	EEByte		; Data byte to write
; Storage for ASCII decimal conversion and digits
	inNum		; Source operand
	thisDig		; Digit counter
	endc

;============================================================
;                        P R O G R A M
;============================================================
		org		0	  ; start at address 
		goto	main
; Space for interrupt handlers
	org		0x08
main:
; Wiring:
;     LCD data to port D, lines 0 to 7
;     E line -> port E, 1
;     RW line -> port E, 2
;     RS line -> port E, 0
; Set PORTE D and E for output
; Data memory bank selection bits:
; RP1:RP0          Bank
;   0:0             0    Ports A,B,C,D, and E
;   0:1             1    Tris A,B,C,D, and E
;   1:0             2
;   1:1             3
; First, initialize port B by clearing latches
	clrf	STATUS
	clrf	PORTB
; Select bank 1 to tris port D for output
	Bank1
; Tris port D for output. Port D lines 4 to 7 are wired
; to LCD data lines. Port D lines 0 to 4 are wired to LEDs.
	movlw	B'00000000'
	movwf	TRISD		; and port D
; By default port A lines are analog. To configure them
; as digital code must set bits 1 and 2 of the ADCON1
; register (in bank 1)
	movlw	0x06		; binary 0000 0110  is code to
						; make all port A lines digial
	movwf	ADCON1
; Port B, lines are wired to keypad swtiches, as follows:
;   7 6 5 4 3 2 1 0
;   | | | | |_|_|_|_____ switch rows (output)
;   |_|_|_|_____________ switch columns (input)
; rows must be defined as output and columns as input
	movlw	b'11110000'
	movwf	TRISB
; Tris port E for output
	movlw	B'00000000'
	movwf	TRISE		; Tris port E
; NJU6355 Interface:
; CLK		PORTC,1		Output
; DAT		PORTC,3		Output
; IO		PORTC,5		Output
; CE		PORTA,2		Output
	movlw	b'00000000'
	movwf	TRISC
	movlw	TRISA
; Enable port B pullups for switches in OPTION register
;   7  6  5  4  3  2  1  0 <= OPTION bits
;   |  |  |  |  |  |__|__|_____ PS2-PS0 (prescaler bits)
;   |  |  |  |  |               Values for Timer0
;   |  |  |  |  |               000 = 1:2   001 = 1:4
;   |  |  |  |  |               010 = 1:8   011 = 1:16
;   |  |  |  |  |               100 = 1:32  101 = 1:64
;   |  |  |  |  |               110 = 1:128 *111 = 1:256
;   |  |  |  |  |______________ PSA (prescaler assign)
;   |  |  |  |                 *1 = to WDT
;   |  |  |  |                  0 = to Timer0
;   |  |  |  |_________________ TOSE (Timer0 edge select)
;   |  |  |                    *0 = increment on low-to-high
;   |  |  |                     1 = increment in high-to-low
;   |  |  |____________________ TOCS (TMR0 clock source)
;   |  |                       *0 = internal clock
;   |  |                        1 = RA4/TOCKI bit source
;   |  |_______________________ INTEDG (Edge select)
;   |                          *0 = falling edge
;   |__________________________ RBPU (Pullup enable)
;                              *0 = enabled
;                               1 = dissabled
	movlw	b'00001000'
	movwf	OPTION_REG
; Clear the write error flag (WRERR) in EECON1 
	Bank3
	bcf		EECON1,WRERR
; Back to bank 0
	Bank0
; Clear all output lines
	movlw	b'00000000' 
	movwf	PORTD
	movwf	PORTE
	movwf	PORTA
	movwf	PORTC
; Wait and initialize HD44780
	call	delay_5		; Allow LCD time to initialize itself
	call	initLCD		; Then do forced initialization
	call	delay_5		; (Wait probably not necessary)
; Clear character counter and line counter variables
	clrf	LCDcount
	clrf	LCDline
; Store base address of text buffer in PIC RAM
	movlw	0x20		; Start address for buffer
	movwf	pic_ad		; to local variable
; Initialize EEPROM address and data
	clrf	EEMemAdd	; Set address to 0
	clrf	EEByte
;======================
;   first LCD line
;======================
; Store 20 blanks in PIC RAM, starting at address stored
; in variable pic_ad
	call	blank20
; Call procedure to store ASCII characters for message
; in text buffer
	movlw	d'0'		; Offset into buffer
	call	storeMS1
;=======================
;  Read EEPROM memory
;=======================
; EEPROM memory address to use is at 10 (0x0a). Variable
; EEMemAdd is already initialized.
; Fill data for EEPROM is 0xff. This value indicates
; the first iteration  
	call 	EERead		; Local procedure. Value in w
	movwf	EEByte		; Store value
; At this point w must be 0
; EEPROM data still in w
	clrf	EEMemAdd		; Address 0
	incf	EEByte,f
	call	EEWrite
; At this point iteration number is stored in EEByte
; This value must be displayed on the LCD at offset 11
; of the first line. This means it must be stored at offset
; 11 in the buffer. Since the buffer starts at 0x20 the
; iteration digit must be stored at offset 0x20+11=0x2b 
ShowEEData:
; Binary data in EEByte 
	movf	EEByte,w	; Value to w
	call	bin2asc		; Conversion routine
; At this point three ASCII digits are stored in local
; variables. Move digits to display area
	movf	asc1,w		; Unit digit
	movwf	0x2b		; Store in buffer
	movf	asc10,w		; same with other digits
	movwf	0x2a
	movf	asc100,w
	movwf	0x29
; Display line
; Set DDRAM address to start of first line

showLine:

; Testing real time clock
	call	initRTC		; Initialize variables
	call	Set_Time

; Wait
	movlw	.50
	movwf	count3
longWait:
	call	long_delay
	decfsz	count3,f
	goto	longWait

; Get variables from RTC
	call	Get_Time
over1:
; Store data in EEPROM
	clrf	EEMemAdd		; Address 0
	incf	EEMemAdd,f		; Address 1
;	movf	Hours,w
	movlw	0xee
	movwf	EEByte
	call	EEWrite

	incf	EEMemAdd,f		; Address 2
	movf	Minutes,w
	movwf	EEByte
	call	EEWrite

	incf	EEMemAdd,f		; Address 3
	movf	Seconds,w
	movwf	EEByte
	call	EEWrite

	call 	line1
; Call procedure to display 16 characters in LCD
	call	display20 
loopHere:
	goto    loopHere  ;done

;============================================================
;============================================================
;                L O C A L    P R O C E D U R E S 
;============================================================
;============================================================
;==========================
; init LCD for 4-bit mode 
;==========================
initLCD:
; Initialization for Densitron LCD module as follows:
;	4-bit interface
;   2 display lines of 16 characters each
;   cursor on
;   left-to-right increment
;   cursor shift right
;   no display shift
;=======================|
;   set command mode    |
;=======================|
	bcf		PORTE,E_line	; E line low
	bcf		PORTE,RS_line	; RS line low
	bcf		PORTE,RW_line	; Write mode
	call	delay_125		; delay 125 microseconds
;***********************|
;     FUNCTION SET      |
;***********************|
	movlw	0x28	; 0 0 1 0 1 0 0 0 (FUNCTION SET)
					;     | | | |__ font select:
					;     | | |    1 = 5x10 in 1/8 or 1/11 dc
					;     | | |    0 = 1/16 dc
					;     | | |___ Duty cycle select
					;     | |      0 = 1/8 or 1/11
					;     | |      1 = 1/16 (multiple lines)
					;     | |___ Interface width
					;     |      0 = 4 bits
					;     |      1 = 8 bits
					;     |___ FUNCTION SET COMMAND 
	call	send8	; 4-bit send routine

; Set 4-bit mode command must be repeated
	movlw	0x28
	call 	send8

;***********************|
; DISPLAY AND CURSOR ON |
;***********************|
	movlw	0x0e	; 0 0 0 0 1 1 1 0 (DISPLAY ON/OFF)
					;         | | | |___ Blink character at cursor
					;         | | |      1 = on, 0 = off
					;         | | |___ Curson on/off
					;         | |      1 = on, 0 = off
					;         | |____ Display on/off
					;         |       1 = on, 0 = off
					;         |____ COMMAND BIT
	call	send8
;***********************|
;   set entry mode      |
;***********************|
	movlw	0x06	; 0 0 0 0 0 1 1 0 (ENTRY MODE SET)
					;           | | |___ display shift
					;           | |      1 = shift
					;           | |      0 = no shift
					;           | |____ cursor increment mode
					;           |       1 = left-to-right
					;           |       0 = right-to-left
					;           |___ COMMAND BIT
	call	send8

;***********************|
; cursor/display shift  |
;***********************|
	movlw	0x14	; 0 0 0 1 0 1 0 0 (CURSOR/DISPLAY SHIFT)
   					;       | | | |_|___ don't care
					;       | |_|__ cursor/display shift
					;       |       00 = cursor shift left
					;       |       01 = cursor shift right
					;       |       10 = cursor and display
					;       |            shifted left
					;       |       11 = cursor and display
					;       |            shifted right
					;       |___ COMMAND BIT
	call	send8
;***********************|
;    clear display      |
;***********************|
	movlw	0x01	; 0 0 0 0 0 0 0 1 (CLEAR DISPLAY)
					;               |___ COMMAND BIT
	call	send8
; Per documentation
	call	delay_5	; Test for busy
	return

;=======================
;  Procedure to delay
;   42 microseconds
;=======================
delay_125:
	movlw	D'42'			; Repeat 42 machine cycles
	movwf	count1			; Store value in counter
repeat
	decfsz	count1,f   		; Decrement counter
	goto	repeat			; Continue if not 0
	return					; End of delay	

;=======================
;  Procedure to delay
;   5 milliseconds
;=======================
delay_5:
	movlw	D'42'			; Counter = 41
	movwf	count2			; Store in variable
delay
	call	delay_125		; Delay
	decfsz	count2,f    	; 40 times = 5 milliseconds
	goto	delay
	return				; End of delay
;========================
;     pulse E line 
;========================
pulseE
	bsf		PORTE,E_line	; Pulse E line
	nop
	bcf		PORTE,E_line
	return

;=============================
;   long delay sub-routine
;=============================
long_delay
		movlw	D'200'	; w delay count
		movwf	J		; J = w
jloop:	movwf	K		; K = w
kloop:	decfsz	K,f		; K = K-1, skip next if zero
		goto 	kloop
		decfsz	J,f		; J = J-1, skip next if zero
		goto	jloop
		return

;=============================
;   LCD display procedure
;=============================
; Sends 20 characters from PIC buffer with address stored
; in variable pic_ad to LCD line previously selected
display20:
	call	delay_5			; Make sure not busy
; Set up for data
	bcf		PORTA,E_line	; E line low
	bsf		PORTA,RS_line	; RS line high for data
; Set up counter for 20 characters
	movlw	D'20'
	movwf	count3	
; Get display address from local variable pic_ad
	movf	pic_ad,w		; First display RAM address to W
	movwf	FSR				; W to FSR
getchar
	movf	INDF,w			; get character from display RAM
							; location pointed to by file select
							; register
	call	send8			; 4-bit interface routine
; Test for 16 characters displayed
	decfsz	count3,f  		; Decrement counter
	goto	nextchar		; Skipped if done
	return
nextchar:
	incf	FSR,f			; Bump pointer
	goto	getchar


;========================
;   send 2 nibbles in
;     4-bit mode
;========================
; Procedure to send two 4-bit values to port B lines
; 7, 6, 5, and 4. High-order nibble is sent first
; ON ENTRY:
;         w register holds 8-bit value to send
send8:
	movwf	store1		; Save original value
	call	merge4		; Merge with port B
; Now w has merged byte
	movwf	PORTD		; w to port D
	call	pulseE		; Send data to LCD
; High nibble is sent
	movf	store1,w	; Recover byte into w
	swapf	store1,w	; Swap nibbles in w
	call	merge4
	movwf	PORTD
	call	pulseE		; Send data to LCD
	call	delay_125
	return
;==========================
;       merge bits
;==========================
; Routine to merge the 4 high-order bits of the
; value to send with the contents of port B
; so as to preserve the 4 low-bits in port B
; Logic:
;      AND value with 1111 0000 mask
;      AND port B with 0000 1111 mask
;      Now low nibble in value and high nibble in
;      port B are al 0 bits:
;          value = vvvv 0000
;         port B = 0000 bbbb
;      OR value and port B resulting in:
;                  vvvv bbbb 
; ON ENTRY:
;      w contain value bits
; ON EXIT:
;      w contains merged bits
merge4:
	andlw	b'11110000'	; ANDing with 0 clears the
						; bit. ANDing with 1 preserves
						; the original value
	movwf	store2		; Save result in variable
	movf	PORTD,w		; port B to w register
	andlw	b'00001111' ; Clear high nibble in port b
						; and preserve low nibble
	iorwf	store2,w	; OR two operands in w
	return

;==========================
;   Set address register
;      to LCD line 2
;==========================
; ON ENTRY:
;         Address of LCD line 2 in constant LCD_2 
line2:
	bcf		PORTE,E_line	; E line low
	bcf		PORTE,RS_line	; RS line low, setup for control
	call	delay_5		; Busy?
; Set to second display line
	movlw	LCD_2			; Address with high-bit set
	call	send8
; Set RS line for data
	bsf		PORTE,RS_line	; RS = 1 for data
	call	delay_5			; Busy?
	return

;==========================
;   Set address register
;      to LCD line 1
;==========================
; ON ENTRY:
;         Address of LCD line 1 in constant LCD_1 
line1:
	bcf		PORTE,E_line 	; E line low
	bcf		PORTE,RS_line	; RS line low, set up for control
	call	delay_5		; busy?
; Set to second display line
	movlw	LCD_1			; Address and command bit
	call	send8			; 4-bit routine
; Set RS line for data
	bsf		PORTE,RS_line	; Setup for data
	call	delay_5			; Busy?
	return

;==========================
;   scroll to LCD line 2 
;==========================
; Procedure to count the number of characters displayed on
; each LCD line. If the number reaches the value in the
; constant LCDlimit, then display is scrolled to the second
; LCD line. If at the end of the second line, then LCD is
; reset to the first line.
LCDscroll:
	incf	LCDcount,f		; Bump counter
; Test for line limit
	movf	LCDcount,w
	sublw	LCDlimit		; Count minus limit
	btfss	STATUS,Z		; Is count - limit = 0
	goto	scrollExit		; Go if not at end of line
; At this point the end of the LCD line was reached
; Test if this is also the end of the second line
	movf	LCDline,w
	sublw	0x01			; Is it line 1?
	btfsc	STATUS,Z		; Is LCDline minus 1 = 0?
	goto	line2End		; Go if end of second line
; At this point it is the end of the top LCD line
	call	line2			; Scroll to second line
	clrf	LCDcount		; Reset counter
	incf	LCDline,f		; Bump line counter
	goto	scrollExit
; End of second LCD line
line2End:
	call	initLCD			; Reset
	clrf	LCDcount		; Clear counters
	clrf	LCDline
	call	line1			; Display to first line
scrollExit:
	return

;===============================
;  first text string procedure
;===============================
storeMS1:
; Procedure to store in PIC RAM buffer the message
; contained in the code area labeled msg1
; ON ENTRY:
;         variable pic_ad holds address of text buffer
;         in PIC RAM
;         w register hold offset into storage area
;         msg1 is routine that returns the string characters
;         an a zero terminator
;         index is local variable that hold offset into
;         text table. This variable is also used for
;         temporary storage of offset into buffer 
; ON EXIT:
;         Text message stored in buffer
;
; Store offset into text buffer (passed in the w register)
; in temporary variable 
	movwf	index		; Store w in index
; Store base address of text buffer in FSR
	movf	pic_ad,w	; first display RAM address to W
	addwf	index,w		; Add offset to address
	movwf	FSR			; W to FSR
; Initialize index for text string access
	movlw	0			; Start at 0
	movwf	index		; Store index in variable
; w still = 0
get_msg_char:
	call	msg1		; Get character from table
; Test for zero terminator
	andlw	0x0ff
	btfsc	STATUS,Z	; Test zero flag
	goto	endstr1		; End of string
; ASSERT: valid string character in w
;         store character in text buffer (by FSR)
	movwf	INDF		; store in buffer by FSR
	incf	FSR,f		; increment buffer pointer
; Restore table character counter from variable
	movf	index,w		; Get value into w
	addlw	1			; Bump to next character
	movwf	index		; Store table index in variable
	goto	get_msg_char	; Continue	
endstr1:
	return
; Routine for returning message stored in program area
; Message has 10 characters
msg1:
	addwf	PCL,f		; Access table
	retlw	'I'
	retlw	't'
	retlw	'e'
	retlw	'r'
	retlw	'.'
	retlw	0x20
	retlw	'N'
	retlw	'o'
	retlw	'.'
	retlw	0x20
	retlw	0
;========================
;     blank buffer
;========================
; Procedure to store 20 blank characters in PIC RAM
; buffer starting at address stored in the variable
; pic_ad
blank20:
	movlw	D'20'		; Setup counter
	movwf	count1
	movf	pic_ad,w	; First PIC RAM address
	movwf	FSR			; Indexed addressing
	movlw	0x20		; ASCII space character
storeit
	movwf	INDF		; Store blank character in PIC RAM
						; buffer using FSR register
	decfsz	count1,f  	; Done?
	goto	incfsr		; no
	return				; yes
incfsr
	incf	FSR,f		; Bump FSR to next buffer space
	goto	storeit

;==============================
;   binary to ASCII decimal
;        conversion
;============================== 
; ON ENTRY:
;         w register has binary value in range 0 to 255
; ON EXIT:
;         output variables asc100, asc10, and asc1 have
;         three ASCII decimal digits
; Routine logic:
;   The value 100 is subtracted from the source operand
;   until the reminder is < 0 (carry cleared). The number
;   of subtractions is the decimal hundreds result. 100 is
;   then added back to the subtrahend to compensate
;   for the last subtraction. Now 10 is subracted in the
;   same manner to determine the decimal tenths result.
;   The final reminder is the decimal units result.
; Variables:
;     inNum     storage for source operand
;     asc100    storage for hundreds position result
;     asc10     storage for tenth position result
;     asc1      storage for unit position reslt
;     thisDig	Digit counter 
bin2asc:
	movwf	inNum	; Save copy of source value
	clrf	asc100  ; Clear hundreds storage
	clrf	asc10	; Tens
	clrf	asc1	; Units
	clrf	thisDig
sub100:
	movlw	.100
	subwf	inNum,f   	; Subtract 100
	btfsc	STATUS,C   	; Did subtract overflow?
	goto	bump100		; No. Count subtraction
	goto	end100
bump100:
	incf	thisDig,f	;increment digit counter
	goto	sub100
; Store 100th digit
end100:
	movf	thisDig,w  	; Adjusted digit counter
	addlw	0x30		; Convert to ASCII
	movwf	asc100		; Store it
; Calculate tenth position value
	clrf	thisDig
; Adjust minuend
	movlw	.100		; Minuend
	addwf	inNum,f		; Add value to minuend to 
						; compensate for last operation
sub10:
	movlw	.10
	subwf	inNum,f   	; Subtract 10
	btfsc	STATUS,C   	; Did subtract overflow?
	goto	bump10		; No. Count subtraction
	goto	end10
bump10:
	incf	thisDig,f	;increment digit counter
	goto	sub10
; Store 10th digit
end10:
	movlw	.10
	addwf	inNum,f		; Adjust for last subratct
	movf	thisDig,w  	; get digit counter contents
	addlw	0x30		; Conver to ASCII
	movwf	asc10		; Store it
; Calculate and store units digit
	movf	inNum,w   ;	Store units value
	addlw	0x30		; Conver to ASCII
	movwf	asc1		; Store digit
	return
;============================================================
;                 local EEPROM data procedures
;============================================================
; GPRs used in EEPROM-related code are placed in the common
; RAM area (from 0x70 to 0x7f). This makes the registers
; accessible from any bank. 
;==============================
;     read local EEPROM
;==============================
; Procedure to read EEPROM memory
; ON ENTRY:
; Address of EEPROM memory location to read is stored in
; local register EEMemAdd
; ON EXIT:
; Read data in w
EERead:
	Bank2
	movf 	EEMemAdd,W 	; EEPROM address
	movwf 	EEADR 		; to read from
	Bank3
	bcf 	EECON1,EEPGD	; Point to Data memory
	bsf 	EECON1,RD	; Start read
	Bank2
	movf 	EEDATA,W 	; Data to w register
	Bank0
	return

;==============================
;     write local EEPROM
;==============================
; Procedure to write data byte to EEPROM memory
; ON ENTRY:
; Address to write stored in local register EEMemAdd
; Data byte to write is in local register EEByte
EEWrite:
	Bank3
Wait2Start:
	btfsc 	EECON1,WR 		; Wait for
	GOTO 	Wait2Start		; write to finish
	Bank2
	movf 	EEMemAdd,w 		; Address to
	movwf 	EEADR 			; SFR
	movf 	EEByte,w 		; Data to
	movwf 	EEDATA 			; SFR
	Bank3
	bcf 	EECON1,EEPGD 	; Point to Data memory
	bsf 	EECON1,WREN 	; and enable writes
; Dissable interrupts. Can be done in any case
	bcf 	INTCON,GIE
; Write special codes
	movlw 	0x55 			; First code is 0x55
	movwf 	EECON2
	movlw	0xaa 			; Second code is 0xaa
	movwf 	EECON2
	bsf 	EECON1,WR 		; Start write operation
	nop						; Time for write
	nop
; Test for end of write operation
wait2End:
	btfsc	EECON1,WR		; Wait until WR clear
	goto	wait2End
;
; Reenable interrupts if program uses interrupts
; If not, comment out next line
;	bsf 	INTCON,GIE
;
	bcf 	EECON1,WREN 	; Prevent accidental writes
	Bank0
	return
;============================================================
;                    6355 RTC procedures
;============================================================
;------ This routine sets the time for the RTC.  Note that upon
; power-up, the RTC has NO TIME SET.  You must define the time and
; set it in the RTC before attempting to read it. Otherwise, no
; values will be returned.
;------ Note: the Get_Time routine reads the variables from the
; RTC and stores them in BCD format, not Hex.  If you want to
; "reset" the time with the Set_Time routine, then you don't
; have to use the Hex2BCD conversion.
;------ In fact, if the original variables are defined in BCD
; format, you don't need to convert, either.  But, since Hex is
; easier to work with and necessary for table lookup, it might
; be worthwhile to convert back and forth if you need to
; manipulate the variables or use a table.

Set_Time
; NJU6355 Interface for setting time:
; CLK		PORTC,1		Output
; DAT		PORTC,3		Output
; IO		PORTC,5		Output
; CE		PORTA,2		Output
	Bank1
	movlw	b'00000000'
	movwf	TRISC
	movlw	TRISA
	Bank0

	bcf	CLK	  	; need CLK low for CE toggle
	call	delay_5	
	bsf	IO		; see timing diagram from NJU6355 
				; datasheet
	call	delay_5
	bsf	CE
; Send 44 bits total
;(Year[8],Month[8],Date[8],Day[4],Hours[8],Minutes[8])
	movf	Year,w
	call	Hex2BCD		; must convert each variable to
                        ; BCD before sending to RTC
	movwf	Time_Temp
	call	Write_RTC

	movf	Month,w
	call	Hex2BCD
	movwf	Time_Temp
	call	Write_RTC

	movf	Date,w
	call	Hex2BCD
	movwf	Time_Temp
	call	Write_RTC

	movf	Day,w
	call	Hex2BCD
	movwf	Time_Temp
	call	Write_RTC_4bits	; Day input is only 4 bits

	movf	Hours,w
	call	Hex2BCD
	movwf	Time_Temp
	call	Write_RTC

	movf	Minutes,w
	call	Hex2BCD
	movwf	Time_Temp
	call	Write_RTC

; Ignore Seconds - RTC does not accept "seconds" input
	bcf		CLK		; finally pull the clock low
                                ; for CE toggle
	call	delay_5
	bcf		CE		; send the final write command
                                ; to the RTC
	call	delay_5
	bcf		IO		; leave the RTC in output mode
	return

;------ This routine grabs the current time from the RTC and
; STORES IT IN BCD FORMAT in the registers Year, Month, Date,
; Day, Hours, Minutes, Seconds. This is worth noting again: THE
; VARIABLES ARE STORED IN BCD FORMAT, NOT HEX
; NJU6355 Interface for setting time:
; CLK		PORTC,1		Output
; DAT		PORTC,3		Input
; IO		PORTC,5		Output
; CE		PORTA,2		Output

Get_Time
	Bank1
	movlw	b'00001000'
	movwf	TRISC
	movlw	b'00000000'
	movlw	TRISA
	Bank0

	bcf		CLK
	call	delay_5		
	bcf		IO			; see timing diagram from
                        ; NJU6355 datasheet
	call	delay_5
	bsf		CE
; Read all 52 bits ; ;(Year[8],Month[8],Date[8],Day[4],Hours[8],Minutes[8],Seconds[8])
	call	Read_RTC
	movwf	Year

	call	Read_RTC
	movwf	Month

	call	Read_RTC
	movwf	Date

	call	Read_RTC_4bits		; Day value is only 4 bits
	movwf	Day

	call	Read_RTC
	movwf	Hours

	call	Read_RTC
	movwf	Minutes

	call	Read_RTC
	movwf	Seconds

; Turn off RTC output
	bcf		CE
	return

;------ This just clocks the next RTC function into Time_Temp
; and returns it in W.
; Call with: nothing required for call
; Returns with: time function byte in W
; Uses: Time_Temp as temporary storage and Time_Count for bit
; counter

Read_RTC_4bits
	movlw	0x04
	goto	$+2
Read_RTC
	movlw	0x08
	movwf	Time_Count
Read_Loop
	bsf		CLK		; data valid after CLK->high
	bsf		STATUS, C	; default is "1" output from RTC
	btfss	DAT
	bcf		STATUS, C
	bcf		CLK
	rrf		Time_Temp, 1	; LSB output first
	decfsz	Time_Count, 1
	goto	Read_Loop
	movf	Time_Temp, 0	; move into W for return
	return

;------ This just clocks whatever is in the W register out to
; the RTC. Use "Write_RTC" to send 8-bit values.  Use
; "Write_RTC_4bits" for 4-bit values.
; Call with: number to be sent in Time_Temp
; Returns with: nothing returned
; Uses: Time_Temp as temporary storage

Write_RTC_4bits
	movlw	0x04
	goto	$+2
Write_RTC
	movlw	0x08
	movwf	Time_Count
Write_Loop
	bcf		CLK
	call	delay_5
	bsf		DAT		; default is "1" output from RTC
	btfss	Time_Temp, 0	; send LSB first each time
	bcf		DAT
	call	delay_5
	bsf		CLK		; data valid after CLK->high
	rrf		Time_Temp, 1
	decfsz	Time_Count, 1
	goto	Write_Loop
	return

;------ Must convert from Hex to BCD for output to RTC
; Note that this BCD conversion is limited to input values less
; than 99 decimal (0x63 hex)
; Call with: number to be converted in W
; Returns with: converted BCD number in W
; Uses: Time_Temp and Time_Count to hold value during calculations

Hex2BCD
	movwf	Time_Count	; use as temporary holding
                                ; register for conversion
	clrw
	btfsc	Time_Count, 4
	addlw	0x16
	btfsc	Time_Count, 5
	addlw	0x32
	addlw	0x06
	skpdc
	addlw	-0x06
	btfsc	Time_Count, 6
	addlw	0x64
	addlw	0x06
	skpdc
	addlw	-0x06
	movwf	Time_Temp
	movf	Time_Count, 0
	andlw	0x0F
	addwf	Time_Temp, 0
	skpndc
	addlw	0x06
	addlw	0x06
	skpdc
	addlw	-0x06
	return

;============================
;   init time variables
;============================
; Procedure to initialize time variables for testing
initRTC:
	movlw	.1
	movwf	Year
	movlw	.2
	movwf	Month
	movlw	.3
	movwf	Date
	movlw	.4
	movwf	Day
	movlw	.5
	movwf	Hours
	movlw	.6
	movwf	Minutes
	movlw	.7
	movwf	Seconds
	return
	


;============================================================
	end         ; END OF PROGRAM 
;============================================================     

