; File name: Ser2EEP.asm
; Last revision: May 26, 2006
; Author: Julio Sanchez
; PIC: 16F877
;
; Description:
; Receive character data through RS-232 line and store in
; EEPROM data memory. Received characters are echoed on
; the second LCD line. When <Enter> key is detected (code
; 0x0d) the text stored in EEPROM memory is retrieved and
; displayed on the LCD. On startup the top LCD line displays
; the prompt: "Receiving:". At that time a message "Rdy- " is
; sent through the serial line so as to test the connection.
;
; Default serial line setting:
;              2400 baud
;              no parity
;              1 stop bit
;              8 character bits
;
; Program to uses 4-bit PIC-to-LCD interface.
; Code assumes that LCD is driven by Hitachi HD44780
; controller and PIC 16F977. Display supports two lines
; each one with 20 characters. The length, wiring and base
; address of each display line is stored in #define
; statements. These statements can be edited to acommodate
; a different set-up.
; 
; WARNING:
; Code assumes 10 Mhz clock. Delay routines must be
; edited for a different clock. Clock speed also determines
; values for baud rate setting (see spbrgVal constant).
;
;===========================
;      16F877 switches
;===========================
; Switches used in __config directive:
;   _CP_ON     	  Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF  
;   _BODEN_ON     Brown-out reset enable ON/OFF
; * _BODEN_OFF 
; * _PWRTE_ON     Power-up timer enable ON/OFF
;   _PWRTE_OFF   
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF
;   _LPV_ON       Low voltage IC programming enable ON/OFF
; * _LPV_OFF
;   _CPD_ON       Data EE memory code protection ON/OFF
; * _CPD_OFF
; OSCILLATOR CONFIGURATIONS:        
;   _LP_OSC       Low power crystal oscillator
;   _XT_OSC       External parallel resonator/crystal oscillator  
; * _HS_OSC       High speed crystal resonator
;   _RC_OSC       Resistor/capacitor oscillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

	processor 	16f877		; Define processor
	#include	<p16f877.inc>
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _HS_OSC & _WDT_OFF & _LVP_OFF & _CPD_OFF

; __CONFIG directive is used to embed configuration data
; within the source file. The labels following the directive
; are located in the corresponding .inc file.
	errorlevel -302
; Suppress bank-related warning 
;============================================================
;                          M A C R O S
;============================================================
; Macros to select the register banks
Bank0	MACRO			; Select RAM bank 0
		bcf	STATUS,RP0
		bcf	STATUS,RP1
		ENDM

Bank1	MACRO			; Select RAM bank 1
		bsf	STATUS,RP0
		bcf	STATUS,RP1
		ENDM

Bank2	MACRO			; Select RAM bank 2
		bcf	STATUS,RP0
		bsf	STATUS,RP1
		ENDM

Bank3	MACRO			; Select RAM bank 3
		bsf	STATUS,RP0
		bsf	STATUS,RP1
		ENDM
;=====================================================
;                 constant definitions
;  for PIC-to-LCD pin wiring and LCD line addresses
;=====================================================
#define E_line 1   	;|
#define RS_line 0	;| -- from wiring diagram 
#define RW_line 2	;|
; LCD line addresses (from LCD data sheet)
#define LCD_1 0x80	; First LCD line constant
#define LCD_2 0xc0	; Second LCD line constant
#define LCDlimit .20; Number of characters per line
#define	spbrgVal .64; For 2400 baud on 10Mhz clock
; Note: The constant that define the LCD display
;       line addresses have the high-order bit set
;       so as to meet the requirements of controller
;       commands. 
;
;==========================================================
;                 General Purpose Variables
;==========================================================
; Local variables
; Reserve 20 bytes for string buffer
	cblock	0x20
	strData
	endc
; Other data
	cblock	0x34		; Start of block
	count1		; Counter # 1
	count2		; Counter # 2
	count3		; Counter # 3
	J			; counter J
	K			; counter K
	bufAdd
	index
	store1		; Local storage
	store2
	endc
;==============================
;      Common RAM area
;==============================
; These GPRs can be accessed from any bank.
; 15 bytes are available, from 0x70 to 0x7f
	cblock	0x70
; For LCDscroll procedure
	LCDcount	; Counter for characters per line
	LCDline		; Current display line (0 or 1)
; Communications variables
	newData		; not 0 if new data received
	ascVal	
	errorFlags
; EEPROM-related variables
	EEMemAdd	; EEPROM address to access
	EEByte		; Data byte to write
	endc

;============================================================
;                        P R O G R A M
;============================================================
		org		0	  ; start at address 
		goto	main
; Space for interrupt handlers
	org		0x08
main:
; Wiring:
;     LCD data to port D, lines 0 to 7
;     E line -> port E, 1
;     RW line -> port E, 2
;     RS line -> port E, 0
; Set PORTE D and E for output
; First, initialize port B by clearing latches
	clrf	STATUS
	clrf	PORTB
; Select bank 1 to tris port D for output
	Bank1
; Tris port D for output. Port D lines 4 to 7 are wired
; to LCD data lines. Port D lines 0 to 4 are wired to LEDs.
	movlw	B'00000000'
	movwf	TRISD		; and port D
; By default port A lines are analog. To configure them
; as digital code must set bits 1 and 2 of the ADCON1
; register (in bank 1)
	movlw	0x06		; binary 0000 0110  is code to
						; make all port A lines digital
	movwf	ADCON1
; Port B, lines are wired to keypad switches, as follows:
;   7 6 5 4 3 2 1 0
;   | | | | |_|_|_|_____ switch rows (output)
;   |_|_|_|_____________ switch columns (input)
; rows must be defined as output and columns as input
	movlw	b'11110000'
	movwf	TRISB
; Tris port E for output
	movlw	B'00000000'
	movwf	TRISE		; Tris port E
; Enable port B pullups for switches in OPTION register
	movlw	b'00001000'
	movwf	OPTION_REG
; Back to bank 0
	Bank0
; Initialize serial port for 2400 baud, 8 bits, no parity
; 1 stop
	call	InitSerial
; Test serial transmission by sending "RDY-"
	movlw	'R'
	call	SerialSend
	movlw	'D'
	call	SerialSend
	movlw	'Y'
	call	SerialSend
	movlw	'-'
	call	SerialSend
	movlw	0x20
	call	SerialSend
; Clear all output lines
	movlw	b'00000000' 
	movwf	PORTD
	movwf	PORTE
; Wait and initialize HD44780
	call	delay_5		; Allow LCD time to initialize itself
	call	initLCD		; Then do forced initialization
	call	delay_5		; (Wait probably not necessary)
; Clear character counter and line counter variables
	clrf	LCDcount
	clrf	LCDline
; Set display address to start of first LCD line
	call	line1
; Store address of display buffer
	movlw	0x20
	movwf	bufAdd	
; Display "Receiving:" message prompt
	call	blank20		; Clear buffer
	movlw	0x00		; Offset in buffer
	call	storeMS1	; Store message at offset
	call	display20	; Display message	
; Start address of EEPROM
	clrf	EEMemAdd
; Setup for display in second line
	call	line2
	clrf	LCDline
	incf	LCDline,f		; Set scroll control for line 2	
;============================================================
;           receive serial data, store, and display
;============================================================
receive:
; Call serial receive procedure
	call	SerialRcv
; HOB of newData register is set if new data
; received
	btfss	newData,7
	goto	scanExit
; At this point new data was received.
	movwf	EEByte		; Save received character
; Display character on LCD
	movf	EEByte,w	; Recover character
	call	send8		; Display in LCD
	call	LCDscroll	; Scroll at end of line
; Store character in EEPROM at location in EEMemAdd
	call	EEWrite		; Local procedure
	incf	EEMemAdd,f	; Bump to next EEPROM
; Check for <Enter> key (0x0d) and execute display function
	movf	EEByte,w	; Recover last received
	sublw	0x0d
	btfsc	STATUS,Z	; Test if <Enter> key
	goto	isEnter		; Go if <Enter>
; Not <Enter> key, continue processing
scanExit:
	goto    receive  	; Continue
;============================
;    display EEPROM data
;============================
; This routine receives control when the <Enter> key is 
; received.
; Action:
;        1. Clear LCD
;        2. Output is set to top LCD line
;        3. Characters stored in EEPROM are displayed
;           until 0x0d code is detected
isEnter:
	call	clearLCD
; Clear character counter and line counter variables
	clrf	LCDcount
	clrf	LCDline
; Read data from EEPROM memory, starting at address 0
; and display on LCD until 0x0d terminator
	call	line1
	clrf	EEMemAdd	; Start at EEPROM 0
readOne:
	call	EERead		; Get character
; Store character
	movwf	EEByte		; Save character
; Test for terminator
	sublw	0x0d
	btfsc	STATUS,Z	; Test if 0x0d
	goto	atEnd		; Go if 0x0d
; At this point character read is not 0x0d
; Display on LCD
	movf	EEByte,w	; Recover character	
; Display character on LCD
	call	send8		; Display in LCD
	call	LCDscroll	; Scroll at end of line
	incf	EEMemAdd,f	; Next EEPROM byte
	goto	readOne

; End of execution	
atEnd:
	goto	atEnd

;============================================================
;============================================================
;                L O C A L    P R O C E D U R E S 
;============================================================
;============================================================
;==========================
; init LCD for 4-bit mode 
;==========================
initLCD:
; Initialization for Densitron LCD module as follows:
;	4-bit interface
;   2 display lines of 16 characters each
;   cursor on
;   left-to-right increment
;   cursor shift right
;   no display shift
;=======================|
;   set command mode    |
;=======================|
	bcf		PORTE,E_line	; E line low
	bcf		PORTE,RS_line	; RS line low
	bcf		PORTE,RW_line	; Write mode
	call	delay_125		; delay 125 microseconds
	movlw	0x28	; 0 0 1 0 1 0 0 0 (FUNCTION SET)
	call	send8	; 4-bit send routine
; Set 4-bit mode command must be repeated
	movlw	0x28
	call 	send8
	movlw	0x0e	; 0 0 0 0 1 1 1 0 (DISPLAY ON/OFF)
	call	send8
	movlw	0x06	; 0 0 0 0 0 1 1 0 (ENTRY MODE SET)
	call	send8
	movlw	0x14	; 0 0 0 1 0 1 0 0 (CURSOR/DISPLAY SHIFT)
	call	send8
	movlw	0x01	; 0 0 0 0 0 0 0 1 (CLEAR DISPLAY)
					;               |___ COMMAND BIT
	call	send8
	call	delay_5	; Test for busy
	return

.;===========================
;   procedure to clear LCD
;============================
clearLCD:
	bcf		PORTE,E_line	; E line low
	bcf		PORTE,RS_line	; RS line low
	bcf		PORTE,RW_line	; Write mode
	call	delay_125		; delay 125 microseconds
	movlw	0x01	; 0 0 0 0 0 0 0 1 (CLEAR DISPLAY)
					;               |___ COMMAND BIT
	call	send8
	call	delay_5	; Test for busy
	return

;=======================
;  Procedure to delay
;   42 microseconds
;=======================
delay_125:
	movlw	.105			; Repeat 105 machine cycles
	movwf	count1			; Store value in counter
repeat
	decfsz	count1,f   		; Decrement counter
	goto	repeat			; Continue if not 0
	return					; End of delay	

;=======================
;  Procedure to delay
;   5 milliseconds
;=======================
delay_5:
	movlw	.105			; Counter = 105 cycles
	movwf	count2			; Store in variable
delay
	call	delay_125		; Delay
	decfsz	count2,f    	; 40 times = 5 milliseconds
	goto	delay
	return				; End of delay
;========================
;     pulse E line 
;========================
pulseE
	bsf		PORTE,E_line	; Pulse E line
	nop
	bcf		PORTE,E_line
	return

;=============================
;   long delay sub-routine
;=============================
long_delay
		movlw	D'200'	; w delay count
		movwf	J		; J = w
jloop:	movwf	K		; K = w
kloop:	decfsz	K,f		; K = K-1, skip next if zero
		goto 	kloop
		decfsz	J,f		; J = J-1, skip next if zero
		goto	jloop
		return
;========================
;   send 2 nibbles in
;     4-bit mode
;========================
; Procedure to send two 4-bit values to port B lines
; 7, 6, 5, and 4. High-order nibble is sent first
; ON ENTRY:
;         w register holds 8-bit value to send
send8:
	movwf	store1		; Save original value
	call	merge4		; Merge with port B
; Now w has merged byte
	movwf	PORTD		; w to port D
	call	pulseE		; Send data to LCD
; High nibble is sent
	movf	store1,w	; Recover byte into w
	swapf	store1,w	; Swap nibbles in w
	call	merge4
	movwf	PORTD
	call	pulseE		; Send data to LCD
	call	delay_125
	return
;==========================
;       merge bits
;==========================
; Routine to merge the 4 high-order bits of the
; value to send with the contents of port B
; so as to preserve the 4 low-bits in port B
; Logic:
;      AND value with 1111 0000 mask
;      AND port B with 0000 1111 mask
;      Now low nibble in value and high nibble in
;      port B are al 0 bits:
;          value = vvvv 0000
;         port B = 0000 bbbb
;      OR value and port B resulting in:
;                  vvvv bbbb 
; ON ENTRY:
;      w contain value bits
; ON EXIT:
;      w contains merged bits
merge4:
	andlw	b'11110000'	; ANDing with 0 clears the
						; bit. ANDing with 1 preserves
						; the original value
	movwf	store2		; Save result in variable
	movf	PORTD,w		; port B to w register
	andlw	b'00001111' ; Clear high nibble in port b
						; and preserve low nibble
	iorwf	store2,w	; OR two operands in w
	return
;==========================
;   Set address register
;      to LCD line 2
;==========================
; ON ENTRY:
;         Address of LCD line 2 in constant LCD_2 
line2:
	bcf		PORTE,E_line	; E line low
	bcf		PORTE,RS_line	; RS line low, setup for control
	call	delay_5		; Busy?
; Set to second display line
	movlw	LCD_2			; Address with high-bit set
	call	send8
; Set RS line for data
	bsf		PORTE,RS_line	; RS = 1 for data
	call	delay_5			; Busy?
	return
;==========================
;   Set address register
;      to LCD line 1
;==========================
; ON ENTRY:
;         Address of LCD line 1 in constant LCD_1 
line1:
	bcf		PORTE,E_line 	; E line low
	bcf		PORTE,RS_line	; RS line low, set up for control
	call	delay_5		; busy?
; Set to second display line
	movlw	LCD_1			; Address and command bit
	call	send8			; 4-bit routine
; Set RS line for data
	bsf		PORTE,RS_line	; Setup for data
	call	delay_5			; Busy?
	return

;==========================
;   scroll to LCD line 2 
;==========================
; Procedure to count the number of characters displayed on
; each LCD line. If the number reaches the value in the
; constant LCDlimit, then display is scrolled to the second
; LCD line. If at the end of the second line, then LCD is
; reset to the first line.
LCDscroll:
	incf	LCDcount,f		; Bump counter
; Test for line limit
	movf	LCDcount,w
	sublw	LCDlimit		; Count minus limit
	btfss	STATUS,Z		; Is count - limit = 0
	goto	scrollExit		; Go if not at end of line
; At this point the end of the LCD line was reached
; Test if this is also the end of the second line
	movf	LCDline,w
	sublw	0x01			; Is it line 1?
	btfsc	STATUS,Z		; Is LCDline minus 1 = 0?
	goto	line2End		; Go if end of second line
; At this point it is the end of the top LCD line
	call	line2			; Scroll to second line
	clrf	LCDcount		; Reset counter
	incf	LCDline,f		; Bump line counter
	goto	scrollExit
; End of second LCD line
line2End:
	call	initLCD			; Reset
	clrf	LCDcount		; Clear counters
	clrf	LCDline
	call	line1			; Display to first line
scrollExit:
	return

;=============================
;   LCD display procedure
;=============================
; Sends 20 characters from PIC buffer with address stored
; in variable bufAdd to LCD line previously selected
display20:
	call	delay_5			; Make sure not busy
; Set up for data
	bcf		PORTA,E_line	; E line low
	bsf		PORTA,RS_line	; RS line high for data
; Set up counter for 20 characters
	movlw	D'20'
	movwf	count3	
; Get display address from local variable bufAdd
	movf	bufAdd,w		; First display RAM address to W
	movwf	FSR				; W to FSR
getchar
	movf	INDF,w			; get character from display RAM
							; location pointed to by file select
							; register
	call	send8			; 4-bit interface routine
; Test for 20 characters displayed
	decfsz	count3,f  		; Decrement counter
	goto	nextchar		; Skipped if done
	return
nextchar:
	incf	FSR,f			; Bump pointer
	goto	getchar

;===============================
;  first text string procedure
;===============================
storeMS1:
; Procedure to store in PIC RAM buffer the message
; contained in the code area labeled msg1
; ON ENTRY:
;         variable bufAdd holds address of text buffer
;         in PIC RAM
;         w register hold offset into storage area
;         msg1 is routine that returns the string characters
;         an a zero terminator
;         index is local variable that hold offset into
;         text table. This variable is also used for
;         temporary storage of offset into buffer 
; ON EXIT:
;         Text message stored in buffer
;
; Store offset into text buffer (passed in the w register)
; in temporary variable 
	movwf	index		; Store w in index
; Store base address of text buffer in FSR
	movf	bufAdd,w	; first display RAM address to W
	addwf	index,w		; Add offset to address
	movwf	FSR			; W to FSR
; Initialize index for text string access
	movlw	0			; Start at 0
	movwf	index		; Store index in variable
; w still = 0
get_msg_char:
	call	msg1		; Get character from table
; Test for zero terminator
	andlw	0x0ff
	btfsc	STATUS,Z	; Test zero flag
	goto	endstr1		; End of string
; ASSERT: valid string character in w
;         store character in text buffer (by FSR)
	movwf	INDF		; store in buffer by FSR
	incf	FSR,f		; increment buffer pointer
; Restore table character counter from variable
	movf	index,w		; Get value into w
	addlw	1			; Bump to next character
	movwf	index		; Store table index in variable
	goto	get_msg_char	; Continue	
endstr1:
	return

; Routine for returning message stored in program area
; Message has 10 characters
msg1:
	addwf	PCL,f		; Access table
	retlw	'R'
	retlw	'e'
	retlw	'C'
	retlw	'e'
	retlw	'i'
	retlw	'v'
	retlw	'i'
	retlw	'n'
	retlw	'g'
	retlw	':'
	retlw	0


;========================
;     blank buffer
;========================
; Procedure to store 20 blank characters in PIC RAM
; buffer starting at address stored in the variable
; bufAdd
blank20:
	movlw	D'20'		; Setup counter
	movwf	count1
	movf	bufAdd,w	; First PIC RAM address
	movwf	FSR			; Indexed addressing
	movlw	0x20		; ASCII space character
storeit
	movwf	INDF		; Store blank character in PIC RAM
						; buffer using FSR register
	decfsz	count1,f  	; Done?
	goto	incfsr		; no
	return				; yes
incfsr
	incf	FSR,f		; Bump FSR to next buffer space
	goto	storeit

;==============================================================
;               communications procedures
;==============================================================
; Initiazalize serial port for 2400 baud, 8 bits, no parity,
; 1 stop
InitSerial:
	Bank1					; Macro to select bank1
; Bits 6 and 7 of Port C are multiplexed as TX/CK and RX/DT
; for USART operation. These bits must be set to input in the
; TRISC register
	movlw	b'11000000'		; Bits for TX and RX
	iorwf	TRISC,f			; OR into Trisc register
; The asynchronous baud rate is calculated as follows:
;                      Fosc
;              ABR = -----------
;                     S*(x+1)
; Where x is value in the SPBRG register and S is 64 if the high
; baud rate select bit (BRGH) in the TXSTA control register is
; clear, and 16 if the BRGH bit is set. For setting to 2400 baud
; using a 10Mhs oscillator at a slow baud rate the formula
; is:
; At slow speed (BRGH = 0)
;           10,000,000   10,000,000
;           ---------- = ----------- = 2,403.84 (0.16% error)
;           64*(64+1)      4160          
;
	movlw	spbrgVal		; Value in spbrgVal = 64
	movwf	SPBRG			; Place in baud rate generator
; Setup value: 0010 0000 = 0x20
	movlw	0x20 		; Enable transmission and high baud rate
	movwf	TXSTA
	Bank0				; Bank 0
; Setup value: 1001 0000 = 0x90
	movlw	0x90		; Enable serial port and continuous reception
	movwf	RCSTA
; 
	clrf	errorFlags	; Clear local error flags register
	return
;==============================
;       transmit data
;==============================
; Test for Transmit Register Empty and transmit data in w
SerialSend:
	Bank0				; Select bank 0
	btfss	PIR1,TXIF	; check if transmitter busy
	goto	$-1		; wait until transmitter is not busy
	movwf	TXREG		; and transmit the data
	return

;==============================
;      receive data
;==============================
; Procedure to test line for data received and return value
; in w. Overrun and framing errors are detected and
; remembered in the variable errorFlags, as follows:
;       7  6  5  4  3  2  1  0   <== errorFlags
;       -- not used ----  |  |___ overrun error
;                         |______ framing error
SerialRcv:
	clrf	newData		; Clear new data received register
	Bank0				; Select bank 0
; Bit 5 (RCIF) of the PIR1 Register is clear if the USART
; receive buffer is empty. If so, no data has been received
	btfss	PIR1,RCIF	; Check for received data
	return				; Exit if no data
; At this point data has been received. First eliminate
; possible errors: overrun and framing.
; Bit 1 (OERR) of the RCSTA register detects overrun
; Bit 2 (FERR( of the RCSTA register detects framing error 
	btfsc	RCSTA,OERR	; Test for overrun error
	goto	OverErr		; Error handler
	btfsc	RCSTA,FERR	; Test for framing error
	goto	FrameErr	; Error handler
; At this point no error was detected
; Received data is in the USART RCREG register
	movf	RCREG,w		; get received data
	bsf		newData,7   ; Set bit 7 to indicate new data
; Clear error flags
	clrf	errorFlags
	return
;==========================
;     error handlers
;==========================
OverErr:
	bsf		errorFlags,0	; Bit 0 is overrun error
; Reset system
	bcf		RCSTA,CREN	; Clear continuous receive bit
	bsf		RCSTA,CREN	; Set to re-enable reception
	return
;error because FERR framing error bit is set
;can do special error handling here - this code simply clears
; and continues
FrameErr:
	bsf		errorFlags,1; Bit 1 is framing error
	movf	RCREG,W		; Read and throw away bad data
	return
;============================================================
;                 local EEPROM data procedures
;============================================================
; GPRs used in EEPROM-related code are placed in the common
; RAM area (from 0x70 to 0x7f). This makes the registers
; accessible from any bank. 
;==============================
;     read local EEPROM
;==============================
; Procedure to read EEPROM memory
; ON ENTRY:
; Address of EEPROM memory location to read is stored in
; local register EEMemAdd
; ON EXIT:
; Read data in w
EERead:
	Bank2
	movf 	EEMemAdd,W 	; EEPROM address
	movwf 	EEADR 		; to read from
	Bank3
	bcf 	EECON1,EEPGD	; Point to Data memory
	bsf 	EECON1,RD	; Start read
	Bank2
	movf 	EEDATA,W 	; Data to w register
	Bank0
	return

;==============================
;     write local EEPROM
;==============================
; Procedure to write data byte to EEPROM memory
; ON ENTRY:
; Address to write stored in local register EEMemAdd
; Data byte to write is in local register EEByte
EEWrite:
	Bank3
Wait2Start:
	btfsc 	EECON1,WR 		; Wait for
	GOTO 	Wait2Start		; write to finish
	Bank2
	movf 	EEMemAdd,w 		; Address to
	movwf 	EEADR 			; SFR
	movf 	EEByte,w 		; Data to
	movwf 	EEDATA 			; SFR
	Bank3
	bcf 	EECON1,EEPGD 	; Point to Data memory
	bsf 	EECON1,WREN 	; and enable writes
; Disable interrupts. Can be done in any case
	bcf 	INTCON,GIE
; Write special codes
	movlw 	0x55 			; First code is 0x55
	movwf 	EECON2
	movlw	0xaa 			; Second code is 0xaa
	movwf 	EECON2
	bsf 	EECON1,WR 		; Start write operation
	nop						; Time for write
	nop
; Test for end of write operation
wait2End:
	btfsc	EECON1,WR		; Wait until WR clear
	goto	wait2End
;
; Re-enable interrupts if program uses interrupts
; If not, comment out next line
;	bsf 	INTCON,GIE
;
	bcf 	EECON1,WREN 	; Prevent accidental writes
	Bank0
	return

;============================================================
	end         ; END OF PROGRAM 
;============================================================     

