; File name: Serial6465.asm
; Last update: May 7, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Program to exercise serial communications using a
; PIC 16F84A and two shift registers: a 74HC164, and a
; 74HC165. The 74HC165 inputs 8 lines from a DIP switch
; and transmits settings to PIC through a serial line.
; PIC sends data serially to an 74HC164 which is wired
; to 8 LEDs that display the received data. A total of
; 6 PIC lines are used in interfacing 8 input switches
; to 8 output LEDs.
; Circuit:
; 	  * Port A0 is the serial transmission line which
;       comes from the 74HC165.
;     * Port A1 is wired to the 74HC164 CLOCK pin
;     * Port A2 is wired to the 74HC164 CLEAR pin
;     * 74HC164 output pins 0 to 7 are wired to LEDs.
;     * Port B0 is wired to the 74HC165 Hout line
;     * Port B1 is wired to the 74HC165 CLK line
;     * Port B2 is wired to the 74HC165 load line
;     * A pushbutton swtich is in the 16F84 RESET line
;       and serves to restart the program
; Communications protocol:
;       Communication between PIC and the 74HC164 and 
;       74HC165 is synchronous since the shift registers
;       clock lines serve to shift in and out the data
;       bits. 
;
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

;=========================
; setup and configuration
;=========================
	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF

;============================================================
;                        M A C R O S
;============================================================
; Macros to select the register banks
Bank0	MACRO			; Select RAM bank 0
		bcf		STATUS,RP0
		ENDM

Bank1	MACRO			; Select RAM bank 1
		bsf		STATUS,RP0
		ENDM
; Note: in the case of the 16F84 the bank select macros
;       do not make the code more efficient, however, they
;       do serve to clarify the the bank selection
;       operations.
;=====================================================
;      constant definitions from wiring diagram
;=====================================================
#define clockLN 1   ;|
#define clearLN 2 	;| -- 74HC164 lines 
#define dataLN 0	;|
;
#define clk65LN 1   ;| -- 74HC165 lines
#define loadLN 2    ;|
;=====================================================
;             PIC register and flag equates
;=====================================================
highBit	equ		7		; High order bit
;======================================================
;              variables in PIC RAM
;=====================================================
	cblock	0x0d	; Start of block
	bitCount		; Counter for 8 bits
	dataReg			; Data to send
	workReg			; Work register for bit shifts
	endc
;=========================================================
;                        program
;=========================================================
	org	0	  ; start at address 
	goto	main
; Space for interrupt handlers
	org	0x04

main:
; Port A is all output
	Bank1
	movlw	b'00000000' 
	movwf	TRISA
; Port B line 0 is input, all others are output
	movlw	b'00000001'
	movwf	TRISB	
	Bank0
; Make sure port A line 2 (clear line) is high
	movlw	b'00000100'
	movwf	PORTA
;========================
; read input from 165 IC
;========================
	call	in165			; Local procedure
; dataReg contains input
;===========================
;    call serial output
;       procedure
;===========================   
	call	out164		; Call serial output procedure
;===========================
;        wait forever
;===========================
endloop:
	goto	endloop
;============================================================
;           74HC164 procedure to send serial data
;============================================================
; ON ENTRY:
;         local variable dataReg holds 8-bit value to be
;         transmitted through port labeled serialLN
; OPERATION:
;      1. A local counter (bitCount) is initialized to
;         8 bits
;      2. Code assumes that the first bit is zero by
;         setting the data line low. Then the high-order
;         bit in the data register (dataReg) is tested.
;         If set, the data line is changed to high.
;      3. Bits are shifted in by pulsing the 74HC164
;         clock line (CLK).
;      4. Data bits are then shifted left and the bit
;         counter is tested. If all 8 bits have been sent
;         the procedure returns. 
out164:
; Clear 74HC164 shift register
	bcf		PORTA,clearLN	; 74HC164 CLR clear low
	bsf		PORTA,clearLN	; then high again
; Init counter
	movlw	0x08		; Initialize bit counter
	movwf	bitCount
sendBit:
	bcf		PORTA,dataLN	; Set data line low (assume)
; Using this assumption is possible because the bit is not
; shifted in until the clock line is pulsed.
	btfsc	dataReg,highBit	; test number bit 7
	bsf		PORTA,dataLN 	; Change assumption if set
;=========================
;    pulse clock line
;=========================
; Bits are shifted in by pulsing the 74HC164 CLK line
	bsf		PORTA,clockLN	; CLK high
	bcf		PORTA,clockLN	; CLK low
;=========================
; Rotate data bits left
;=========================
	rlf		dataReg,f  	; Shift left data bits
	decfsz	bitCount,f	; Decrement bit counter
	goto	sendBit		; Repeat if not 8 bits
;=========================
;   end of transmission
;=========================
	return

;============================================================
;     74HC165 procedure to read parallel data and send
;                    serially to PIC
;============================================================
; OPERATION:
;        1. Eight DIP switches are connected to the input
;           ports of an 74HC165 IC. Its output line Hout,
;           and its control lines CLK and load are connected
;           to the PIC's port B lines 0, 1, and 2
;           respectively
;        2. Procedure sets a counter (bitCount) for 8
;           iterations and clears a data holding register
;           (dataReg).
;        3. Port B bits are read into w. Only the lsb of
;           port B is relevant. Value is stored in a working
;           register and the meaningful bit is rotated into
;           the carry flag, then the carry flag bit is
;           then shifted into the data register.
;        4. The iteration counter is decremented. If this
;           is the last iteration the routine ends. Otherwise
;           the bitwise read-and-write operation is repeated.     
in165:
	clrf	dataReg		; Clear data register
	movlw	0x08		; Initialize counter
	movwf	bitCount
	bcf		PORTB,loadLN	; Reset shift register
	bsf		PORTB,loadLN
nextBit:
	movf	PORTB,w 	; Read port B (only LOB is
                        ; meaningful in this routine)
	movwf	workReg		; Store value in local register
	rrf		workReg,f	; Rotate LOB bit into carry flag
	rlf		dataReg,f   ; Carry flag into dataReg
	decfsz	bitCount,f	; Decrement bit counter
	goto	shiftBits	; Continue if not zero
	return				; done
shiftBits
	bsf		PORTB,clk65LN	; Pulse clock
	bcf		PORTB,clk65LN
	goto	nextBit		; Continue


;=========================================================
;                       end of program
;========================================================
	end


