; File name: SerialRcv.asm
; Date: May 6, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Two programs to exercise serial communications between
; two PIC 16F84A both running at 4Mhs. One program sends
; data through a single line and the other one receives
; it. This program is the receiver.
;
; Circuit:
; 	Port A0 is the serial transmission line 
;       Port A2 is an active-low pushbutton swtich that
;               serves to initiate communications.
;       Port A3 is a LED that is ON when the program is
;               ready to receive data. Once data starts
;               being received the LED is turned OFF.
;       Port B0-B7 are 8 LEDs that display the data bits
;               that have been received.
;       A pushbutton swtich is in the 16F84 RESET line
;               and serves to restart the program
;
; Communications parameters:
;       Timer channel TMR0 is used for synchronizing data
;       transmission. The timer runs at the maximum rate of
;       256 cycles per iteration. In a 4Mhz system the
;       timer rate is 1Mhz, thus the bit rate is
;                       1,000,000/256
;       which is approximately 3,906 microseconds per bit.
;
;       Upon receiving the START bit, the program waits for
;       one half a clock cycle (128 timer beats) to
;       synchornize with the sender.
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

;=========================
; setup and configuration
;=========================
	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF
;============================================================
;                        M A C R O S
;============================================================
; Macros to select the register banks
Bank0	MACRO			; Select RAM bank 0
		bcf		STATUS,RP0
		ENDM

Bank1	MACRO			; Select RAM bank 1
		bsf		STATUS,RP0
		ENDM
;=====================================================
;          constant definitions for pin wiring
;=====================================================
#define readySW  2  ;|
#define readyLED 3	;| -- from wiring diagram 
#define serialLN 0	;|
;=====================================================
;            PIC register and flag equates
;=====================================================
c		equ		0		; Carry flag
tmrOVF	equ		2		; Timer overflow bit
;
;======================================================
;              variables in PIC RAM
;=====================================================
	cblock	0x0c	; Start of block
	bitCount		; Counter for 8 bits
	rcvReg			; Data to send
	temp
	endc
;=========================================================
;                           program
;=========================================================
	org	0	  ; start at address 
	goto	main
; Space for interrupt handlers
	org	0x04

main:
	Bank1
; Port A bits 0 and 2 are input. All others are output
	movlw	b'00000101' ; port A setup
	movwf	TRISA
; Port B is all output
	movlw	b'00000000' ; port B setup
	MOVWF	TRISB
	Bank0
; Turn off all port B LEDs
	clrf	PORTB
; And receiver register
	clrf	rcvReg
; Prepare to set prescaler
	clrf	TMR0
	clrwdt
; Setup OPTION register for full timer speed
	movlw	b'11011000' 
;   1  1  0  1  1  0  0  0 <= OPTION bits
;   |  |  |  |  |  |__|__|_____ PS2-PS0 (prescaler bits)
;   |  |  |  |  |               Values for Timer0
;   |  |  |  |  |              *000 = 1:2   001 = 1:4
;   |  |  |  |  |               010 = 1:8   011 = 1:16
;   |  |  |  |  |               100 = 1:32  101 = 1:64
;   |  |  |  |  |               110 = 1:128 111 = 1:256
;   |  |  |  |  |______________ PSA (prescaler assign)
;   |  |  |  |                 *1 = to WDT
;   |  |  |  |                  0 = to Timer0
;   |  |  |  |_________________ TOSE (Timer0 edge select)
;   |  |  |                     0 = increment on low-to-high
;   |  |  |                    *1 = increment in high-to-low
;   |  |  |____________________ TOCS (TMR0 clock source)
;   |  |                       *0 = internal clock
;   |  |                        1 = RA4/TOCKI bit source
;   |  |_______________________ INTEDG (Edge select)
;   |                           0 = falling edge
;   |                          *1 = raising edge
;   |__________________________ RBPU pullups
;                               0 = enabled
;                              *1 = dissabled
	option
; Dissable interrupts
	bcf		INTCON,5   ; Timer0 overflow dissabled
	bcf		INTCON,7   ; Global interupts dissabled
;=========================
;  wait for READY switch
;     to be pressed
;=========================
ready2rcv:
	btfsc	PORTA,readySW	; Test switch
	goto	ready2rcv		; loop
; Turn ON the ready-to-receive LED
	bsf		PORTA,readyLED
;===========================
;        receiving
;===========================
	call	rcvData		; Call serial input procedure
;===========================
;       data received
;===========================
; Turn ready to receive LED off
	bcf		PORTA,readyLED
; Display received data
	movf	rcvReg,w    ; Byte received to w
	movwf	PORTB		; display in port B
;===========================
;      wait forever
;===========================
endloop:
	goto    endloop
;============================================================
;               procedure to receive serial data
;============================================================
; ON ENTRY:
;         local variable dataReg is used to store 8-bit value
;         received through port (labeled serialLN)
; OPERATION:
;      1. The timer at register TMR0 is set to run at
;         maximum clock speed, that is, 256 clock beats.
;         The timer overflow flag in the INTCON register
;         is set when the timer cycles from 0xff to 0x00.
;      2. When the START signal is received, the code
;         waits for 128 timer beats so as to read data in
;         the middle of the send period.
;      3. Each bit (start, data, and stop bits) is read 
;         at intervals of 256 timer beats.
;      4. The procedure tests the timer ofverflow flag
;         (tmrOVF) to determine when the timer cycle has
;         ended, that is when 256 clock beats have passed.

;------------------------------------------------------------
rcvData:
	clrf	TMR0	; Reset timer
	movlw	0x08	; Initialize bit counter
	movwf	bitCount
;=========================
;   wait for START bit
;=========================
startWait:
	btfsc	PORTA,0		; Is port A0 low?
	goto	startWait	; No. Wait for mark
;=========================
;  offset 128 clock beats
;=========================
; At this point the receiver has found the falling
; edge of the start bit. It must now wait 128 timer
; beats to synchronize in the middle of the sender's
; data rate, as follows:
;             |<========= falling edge of START bit
;             |
;             |-----|<====== 128 clock beats offset
;   ----------.     |     .---------
;             |           |  <== SIGNAL
;              ----------- 
;             |<-- 256 -->|
;
	movlw	0x80		; 128 clock beats offset
	movwf	TMR0		; to TMR0 counter
	bcf		INTCON,tmrOVF	; Clear overflow flag
offsetWait:
	btfss	INTCON,tmrOVF   ; timer overflow?
	goto	offsetWait		; wait until
	btfsc	PORTA,0			; Test start bit for error
	goto	offsetWait		; Recycle if a false start
;==========================
;      receive data
;==========================
	clrf	TMR0		; Restart timer
	bcf		INTCON,tmrOVF	; Clear overflow flag
; Wait for 256 timer cycles for first/next data bit
bitWait:
	btfss	INTCON,tmrOVF  ; Timer cycle end?
	goto	bitWait			; Keep waiting
; Timer has counter 256 beats
	bcf		INTCON,tmrOVF	; Reset overflow flag
	movf	PORTA,w			; Read port A into w
	movwf	temp			; Store value read
	rrf		temp,f			; Rotate bit 0 into carry flag
	rlf		rcvReg,f   		; Rotate carry into rcvReg bit 0
	decfsz	bitCount,f		; 8 bits received
	goto	bitWait			; Next bit
; Wait for one time cycle at end of reception
markWait:
	btfss	INTCON,tmrOVF	; timer overflow flag
	goto	markWait		; keep waiting
;========================
;   end of reception
;========================
	return

;=========================================================
;                       end of program
;=========================================================
	end
