; File: SleepDemo
; Date: April 25, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Program to use the External Interrupt on port RB0
; to terminate the power-down state caused by the 
; SLEEP instruction. A pushbutton switch is connected to
; port RB0. The pushbutton generates the interrupt that
; ends the SLEEP conditions.
; Demonstration:
; A LED on port B, line 1, flashes on and off at 1/2
; second intervals for 20 iterations. At that time the
; program enters the SLEEP condition. Pressing the
; pushbutton switch on line RB0 generates the interrupt
; that ends the SLEEP. 
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator (simplest, 20% error)
; |
; |_____ * indicates setup values

;=========================
; setup and configuration
;=========================
	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF

;=====================================================
;              variables in PIC RAM
;=====================================================
; Local variables
	cblock	0x0d		; Start of block
	J			; counter J
	K			; counter K
	count1		; Auxiliary counter
	count2		; Second auxiliary counter
	old_w		; Context saving 
	old_STATUS  ; Idem
	endc

;========================================================
;                 m a i n   p r o g r a m
;========================================================
	org	0	  ; start at address 0
	goto	main
;
;=============================
;      interrupt handler
;=============================
	org		0x04
	goto	IntServ
;=============================
;       main program
;=============================
main:
; Set up interrupt on falling edge
; by clearing OPTION register bit 6
	movlw	b'10111111'
	option
	movlw	b'11111111'	; Set port a for input
	tris	PORTA
	movlw	b'00000001' ; Port B bit 0 is input
	tris	PORTB		; all others are output
	clrf	PORTB		; All port B to 0
;============================
;      setup interrupts
;============================
; Clear external interrupt flag (INTF = bit 1)
	bcf		INTCON,INTF	; Clear flag
; Enable global interrupts (GIE = bit 7)
; Enable RB0 interrupt (INTE = bit 4) 
	bsf		INTCON,GIE	; Enable global int (bit 7)
	bsf		INTCON,INTE	; Enable RB0 int (bit 4)
;============================
;    flash LED 20 times
;============================
wakeUp:
; Program flashes LED wired to port B, line 2
; 20 times before entering the sleep state
	movlw	D'20'		; Number of iterations
	movwf	count2		; To counter
lights:
	movlw	b'00000010'	; Mask with bit 1 set
	xorwf	PORTB,f		; Complement bit 1 	
	call	long_delay
	call	long_delay
	call	long_delay	
	decfsz	count2,f	; Decrement counter
	goto	lights
; 20 iterations have taken place
	sleep
	nop					; Recommended! 
	goto	wakeUp		; Resume execution

;=======================================================
;           Interrupt Service Routine
;=======================================================
; The interrupt service routine performs no operation
IntServ:
	bcf		INTCON,INTF		; Clear flag
	retfie

;=============================
;   long delay sub-routine
;=============================
long_delay
		movlw	D'200'	; w = 200 decimal
		movwf	J		; J = w
jloop:	movwf	K		; K = w
kloop:	decfsz	K,f		; K = K-1, skip next if zero
		goto 	kloop
		decfsz	J,f		; J = J-1, skip next if zero
		goto	jloop
		return

		end	
