; File name: TTYUsart.asm
; Last update: May, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Description:
; Program to emulate USART operation in PIC code. Uses
; PIC-to-LCD interface. Display has 2 lines, each with
; 16 characters.
; Program operation:
; Characters received from the RS232 line are displayed on
; the LCD. LCD lines scroll automatically. A pushbutton
; activates the send operation by transmiting the text
; string: Ready- which is also displayed on the LCD.
;
; Program communications and LCD parameters are stored in
; #define statements. These statements can be edited to
; accomodate a different set-up. Program uses delay loops
; for interface timing.
;
; WARNING:
; Code assumes 4Mhz clock. Delay routines must be
; edited for faster clock
;
; BAUD RATE CALCULATIONS:
; A 4Mhz clock oscillator has a clock frequency of 1 Mhz:
; Since the baud rate is the number of clock cycles per
; second, for a 4Mhz clock it is:
;              1
; bit time = ------ sec. = 208.33 microseconds
;            4,800 
; Calculating one half the baud rate allows reseting the
; clock from the edge to the center of a time pulse:
;
;          |<======== falling edge of start bit
;          |      |<======== center of bit time
;         >|      |< one-half baud rate
;          |      |
;__________.      |      .____________. 
;          |_____________|            |________  
;            208/2 = 104 
; The PIC clock counts up from 0 to 255. So to implement
; a 104 microsecond delay we must start counting at
; clock beat:
;                255 - 104 = 151
; plus one microsecond for movlw instruction used to
; initialize the clock:
;                151 + 1 = 152
; For one full baud rate delay:
;                255 - 208 = 47 + 1 = 48  
; The following two constants are stored in #define
; statements:
;              halfBaud = 152
;              fullBaud = 48               
; Setting the prescaler to TMR0 reduces the baud rate
; to one-half. Other prescaler values will reduce the
; baud rate accordingly.
;
; Wiring diagram:
;    RB4-RB7 ===> LCD data lines 4 to 7 (output)
;    RB0 =======> MAX202 T2in line (output)
;    RA0 =======> MAX202 R2out line (input)
;    RA1 =======> LCD E line (output)
;    RA2 =======> LCD RS line (output)
;    RA3 =======> LCD R/W line (output - not used)
;    RA4 =======> Pushbutton switch 1
;                 (input - active low)
;
;===========================
;        switches
;===========================
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                 Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator
; |               (simplest, 20% error)
; |
; |_____ * indicates setup values presently selected

;=========================
; setup and configuration
;=========================
	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF

;============================================================
;                        M A C R O S
;============================================================
; Macros to select the register banks
Bank0	MACRO			; Select RAM bank 0
		bcf		STATUS,RP0
		ENDM

Bank1	MACRO			; Select RAM bank 1
		bsf		STATUS,RP0
		ENDM
;============================================================
;                    constant definitions
;      for PIC-to-LCD pin wiring and LCD line addresses
;============================================================
#define E_line 1   	;|
#define RS_line 2	;| -- from wiring diagram 
#define RW_line 3	;|
; LCD line addresses (from LCD data sheet)
#define LCD_1 0x80	; First LCD line constant
#define LCD_2 0xc0	; Second LCD line constant
#define LCDlimit .16; Number of characters per line
; 4800 baud clock countdown values
; Code reduces rate to 2400 baud by entering a minimal
; presclaer to TRM0
#define halfBaud .152 ; For one-half bit time
#define fullBaud .48  ; For one full bit time
;
; Note: The constant that define the LCD display line
;       addresses have the high-order bit set in
;       order to faciliate the controller command 
;
;=====================================================
;              PIC register and flag equates
;=====================================================
z	equ		2	; Zero flag
c	equ		0	; Carry flag
;=====================================================
;           buffer and variables in PIC RAM
;=====================================================
; Create a 16-byte storage area
	cblock	0x0c	; Start of first data block
	lineBuf			; buffer for text storage
	endc
; Leave 16 bytes and Continue with local variables
	cblock	0x1c	; Second data block
	count1		; Counter # 1
	count2		; Counter # 2
	J			; counter J
	K			; counter K
	store1		; Local temporary storage 
	store2		; Storage # 2
; For LCDscroll procedure
	LCDcount	; Counter for characters per line
	LCDline		; Current display line (0 or 1)
	bufPtr		; Buffer pointer
; Variables for serial communications
	tempData	; Temporary storage for bit manipulations
	rcvData		; Final storage for received character
	bitCount	; Bit counter
	sendData	; Character to send
	endc

;=========================================================
;                 m a i n    p r o g r a m 
;=========================================================
		org		0	  ; start at address 
		goto	main
; Space for interrupt handlers
	org		0x08

main:
	Bank1
	movlw	b'00010001' ; Port A lines I/O setup
						; RA0 = RS232 input (R2out)
						; RA4 = Pushbutton SW # 1
	movwf	TRISA
	movlw	b'00000000' ; Port B lines as follow:
;    RB4-RB7 ===> LCD data lines 4 to 7 (output)
;    RB0 =======> MAX202 T2in line (output)                        ; RB0 = 
	movwf	TRISB
	Bank0
; Clear bits in port A output lines
	bcf		PORTA,1		
	bcf		PORTA,2
	bcf		PORTA,3
	movlw	b'00000000' ; All outputs ports low
	movwf	PORTB
; Wait and initialize HD44780
	call	delay_5		; Allow LCD time to initialize
						; itself
	call	delay_5
	call	initLCD		; Then do forced initialization
	call	delay_5		; Wait again
; Set port B, line 0 high so start bit is detected
	bsf		PORTB,0
;============================
;   wait for start command
;============================
; Program waits until pushbutton number 1 is pressed
; to continue execution. Pushbutton 1 is active low.
; and wired to RA4
pb1Wait:
	btfsc	PORTA,4		; Test port A, line 4
	goto	pb1Wait		; Loop if not clear
;============================
;  display and send "Ready-"
;============================
; Set LCD base address
	call	line1
; Initialize system for UART emulation at 2400 baud
	call	initTTY
; Display on LCD and test serial transmission by sending
; the string "Ready-"
	movlw	'R'
	movwf	sendData	; Store in send register
	call	send8		; Local LCD display procedure
	call	sendTTY		; Local send procedure
	movlw	'e'
	movwf	sendData	; Store in send register
	call	send8		; Local LCD display procedure
	call	sendTTY		; Local send procedure
	movlw	'a'
	movwf	sendData	; Store in send register
	call	send8		; Local LCD display procedure
	call	sendTTY		; Local send procedure
	movlw	'd'
	movwf	sendData	; Store in send register
	call	send8		; Local LCD display procedure
	call	sendTTY		; Local send procedure
	movlw	'y'
	movwf	sendData	; Store in send register
	call	send8		; Local LCD display procedure
	call	sendTTY		; Local send procedure
	movlw	'-'
	movwf	sendData	; Store in send register
	call	send8		; Local LCD display procedure
	call	sendTTY		; Local send procedure
; Init  character counter and line counter variables for
; LCD line scroll procedure
	movlw	0x06		; 6 characters already displayed
	movwf	LCDcount	
	clrf	LCDline		; LCD line counter
;============================
;     monitor RS232 line
;============================
nextChar:
	call	rcvTTY			; Receive character
; Store character in local line buffer using indirect
; addressing
; 16-byte buffer named lineBuf starts at address 0x0c
; Register variable bufPtr holds offset into buffer
  	movlw	0x0c			; Buffer base address
	addwf	bufPtr,w		; Add pointer in w
	movwf	FSR				; Value to index register
	movf	rcvData,w		; Character into w
	movwf	INDF			; Store w in [FSR]
	incf	bufPtr,f		; Bump pointer
; Send character (still in w) 
	call	send8			; Display it
	call	LCDscroll		; Scroll display lines
	goto	nextChar		; Continue	


;============================================================
;                initialize LCD for 4-bit mode 
;============================================================
initLCD:
; Initialization for Densitron LCD module as follows:
;	4-bit interface
;   2 display lines of 16 characters each
;   cursor on
;   left-to-right increment
;   cursor shift right
;   no display shift
;=======================|
;   set command mode    |
;=======================|
	bcf		PORTA,E_line	; E line low
	bcf		PORTA,RS_line	; RS line low
	bcf		PORTA,RW_line	; Write mode
	call	delay_125		; delay 125 microseconds
;***********************|
;     FUNCTION SET      |
;***********************|
	movlw	0x28	; 0 0 1 0 1 0 0 0 (FUNCTION SET)
					;     | | | |__ font select:
					;     | | |    1 = 5x10 in 1/8 or 1/11 dc
					;     | | |    0 = 1/16 dc
					;     | | |___ Duty cycle select
					;     | |      0 = 1/8 or 1/11
					;     | |      1 = 1/16 (multiple lines)
					;     | |___ Interface width
					;     |      0 = 4 bits
					;     |      1 = 8 bits
					;     |___ FUNCTION SET COMMAND 
	call	send8	; 4-bit send routine

; Set 4-bit mode command must be repeated
	movlw	0x28
	call 	send8

;***********************|
; DISPLAY AND CURSOR ON |
;***********************|
	movlw	0x0e	; 0 0 0 0 1 1 1 0 (DISPLAY ON/OFF)
					;         | | | |___ Blink character at cursor
					;         | | |      1 = on, 0 = off
					;         | | |___ Curson on/off
					;         | |      1 = on, 0 = off
					;         | |____ Display on/off
					;         |       1 = on, 0 = off
					;         |____ COMMAND BIT
	call	send8
;***********************|
;   set entry mode      |
;***********************|
	movlw	0x06	; 0 0 0 0 0 1 1 0 (ENTRY MODE SET)
					;           | | |___ display shift
					;           | |      1 = shift
					;           | |      0 = no shift
					;           | |____ cursor increment mode
					;           |       1 = left-to-right
					;           |       0 = right-to-left
					;           |___ COMMAND BIT
	call	send8

;***********************|
; cursor/display shift  |
;***********************|
	movlw	0x14	; 0 0 0 1 0 1 0 0 (CURSOR/DISPLAY SHIFT)
   					;       | | | |_|___ don't care
					;       | |_|__ cursor/display shift
					;       |       00 = cursor shift left
					;       |       01 = cursor shift right
					;       |       10 = cursor and display
					;       |            shifted left
					;       |       11 = cursor and display
					;       |            shifted right
					;       |___ COMMAND BIT
	call	send8
;***********************|
;    clear display      |
;***********************|
	movlw	0x01	; 0 0 0 0 0 0 0 1 (CLEAR DISPLAY)
					;               |___ COMMAND BIT
	call	send8
; Per documentation
	call	delay_5	; Test for busy
	return

;=======================
;  Procedure to delay
;   42 microseconds
;=======================
delay_125
	movlw	D'42'			; Repeat 42 machine cycles
	movwf	count1			; Store value in counter
repeat
	decfsz	count1,f   		; Decrement counter
	goto	repeat			; Continue if not 0
	return					; End of delay	

;=======================
;  Procedure to delay
;   5 milliseconds
;=======================
delay_5
	movlw	D'41'			; Counter = 41
	movwf	count2			; Store in variable
delay
	call	delay_125		; Delay
	decfsz	count2,f    	; 40 times = 5 milliseconds
	goto	delay
	return				; End of delay
;========================
;     pulse E line 
;========================
pulseE
	bsf		PORTA,E_line	; Pulse E line
	nop
	bcf		PORTA,E_line
	return

;=============================
;   long delay sub-routine
;     (for debugging)
;=============================
long_delay
		movlw	D'200'	; w = 200 decimal
		movwf	J		; J = w
jloop:	movwf	K		; K = w
kloop:	decfsz	K,f		; K = K-1, skip next if zero
		goto 	kloop
		decfsz	J,f		; J = J-1, skip next if zero
		goto	jloop
		return
;========================
;   send 2 nibbles in
;     4-bit mode
;========================
; Procedure to send two 4-bit values to port B lines
; 7, 6, 5, and 4. High-order nibble is sent first
; ON ENTRY:
;         w register holds 8-bit value to send
send8:
	movwf	store1		; Save original value
	call	merge4		; Merge with port B
; Now w has merged byte
	movwf	PORTB		; w to port B
	call	pulseE		; Send data to LCD
; High nibble is sent
	movf	store1,w	; Recover byte into w
	swapf	store1,w	; Swap nibbles in w
	call	merge4
	movwf	PORTB
	call	pulseE		; Send data to LCD
	call	delay_125 
	return
;=================
;  merge bits
;=================
; Routine to merge the 4 high-order bits of the
; value to send with the contents of port B
; so as to preserve the 4 low-bits in port B
; Logic:
;      AND value with 1111 0000 mask
;      AND port B with 0000 1111 mask
;      Now low nibble in value and high nibble in
;      port B are al 0 bits:
;          value = vvvv 0000
;         port B = 0000 bbbb
;      OR value and port B resulting in:
;                  vvvv bbbb 
; ON ENTRY:
;      w contain value bits
; ON EXIT:
;      w contains merged bits
merge4:
	andlw	b'11110000'	; ANDing with 0 clears the
						; bit. ANDing with 1 preserves
						; the original value
	movwf	store2		; Save result in variable
	movf	PORTB,w		; port B to w register
	andlw	b'00001111' ; Clear high nibble in port b
						; and preserve low nibble
	iorwf	store2,w	; OR two operands in w
	return


;========================
; Set address register
;    to LCD line 1
;========================
; ON ENTRY:
;         Address of LCD line 1 in constant LCD_1 
line1:
	bcf		PORTA,E_line 	; E line low
	bcf		PORTA,RS_line	; RS line low, set up for control
	call	delay_5		; busy?
; Set to second display line
	movlw	LCD_1			; Address and command bit
	call	send8			; 4-bit routine
; Set RS line for data
	bsf		PORTA,RS_line	; Setup for data
	call	delay_5			; Busy?
; Clear buffer and pointer
	call	blankBuf
	clrf	bufPtr			; Clear 
	return
;========================
; Set address register
;    to LCD line 2
;========================
; ON ENTRY:
;         Address of LCD line 2 in constant LCD_2 
line2:
	bcf		PORTA,E_line	; E line low
	bcf		PORTA,RS_line	; RS line low, setup for control
	call	delay_5		; Busy?
; Set to second display line
	movlw	LCD_2			; Address with high-bit set
	call	send8
; Set RS line for data
	bsf		PORTA,RS_line	; RS = 1 for data
	call	delay_5			; Busy?
; Clear buffer and pointer
	call	blankBuf		
	clrf	bufPtr
	return

;==========================
;   scroll LCD line 2 
;==========================
; Procedure to count the number of characters displayed on
; each LCD line. If the number reaches the value in the
; constant LCDlimit, then display is scrolled to the second
; LCD line. If at the end of the second line, then the
; second line is scrolled to the first line and display
; continues at the start of the second line
; reset to the first line.
LCDscroll:
	incf	LCDcount,f		; Bump counter
; Test for line limit
	movf	LCDcount,w
	sublw	LCDlimit		; Count minus limit
	btfss	STATUS,z		; Is count - limit = 0
	goto	scrollExit		; Go if not at end of line
; At this point the end of the LCD line was reached
; Test if this is also the end of the second line
	movf	LCDline,w
	sublw	0x01			; Is it line 1?
	btfsc	STATUS,z		; Is LCDline minus 1 = 0?
	goto	line2End		; Go if end of second line
; At this point it is the end of the top LCD line
	call	line2			; Scroll to second line
	clrf	LCDcount		; Reset counter
	incf	LCDline,f		; Bump line counter
	goto	scrollExit
; End of second LCD line
line2End:
; Scroll second line to first line. Characters to be
; scrolled are stored in buffer starting at address 0x0c.
; 16 characters are to be moved
; First clear LCD
	call	initLCD
	call	delay_5			; Make sure not busy
; Set up for data
	bcf		PORTA,E_line	; E line low
	bsf		PORTA,RS_line	; RS line high for data
; Set up counter for 16 characters
	movlw	D'16'			; Counter = 16
	movwf	count2	
; Get address of storage buffer
	movlw	0x0c
	movwf	FSR				; W to FSR
getchar:
	movf	INDF,w			; get character from display RAM
							; location pointed to by file select
							; register
	call	send8			; 4-bit interface routine
; Test for 16 characters displayed
	decfsz	count2,f  		; Decrement counter
	goto	nextchar		; Skipped if done
; At this point scroll operation has concluded
	clrf	LCDcount		; Clear counters
; Stay at line 2
	clrf	LCDline
	incf	LCDline,f
	call	line2			; Set for second line
scrollExit:
	return
nextchar:
	incf	FSR,f			; Bump pointer
	goto	getchar

;============================
;   clear line buffer
;============================
; Use indirect addressing to store 16 blanks in the
; buffer located at 0x0c 
blankBuf:
	Bank0
	movlw	0x0c		; Pointer to RAM
	movwf	FSR			; To index register
blank16:
	clrf	INDF		; Clear memoty pointed at by FSR
	incf	FSR,f		; Bump pointer
	btfss	FSR,4		; 000x0000 when bit 4 is set
						; count reached 16
	goto	blank16
	return
		 
;============================================================
;                    initialize for TTY
;============================================================
; Procedure to initialize RS232 reception 
; Assumes:
;			2400 baud
;			8 data bits
;			no parity
; 			one stop bit
initTTY:
; First initialize receiver to RS-232 line parameters
; Disable global and peripheral interrupts
;  7  6  5  4  3  2  1  0  <= INTCON bitmap
;  |  ?  |  ?  ?  ?  ?  ?  (? = unrelated bits)
;  |     |________________ Timer0 interrupt on overflow
;  |______________________ Global interrupts
	bcf		INTCON,5   	; Disable TMR0 interrupts
	bcf		INTCON,7   	; Dissable global interrupts
	clrf	TMR0		; Reset timer
	clrwdt				; Clear WDT for prescaler assign
	Bank1
; Set up the OPTION regiser bit map
;   7  6  5  4  3  2  1  0 <= OPTION bits
;   1  1  0  1  1  0  0  0 <= setup
;   |  |  |  |  |  |__|__|_____ PS2-PS0 (prescaler bits)
;   |  |  |  |  |               Values for Timer0
;   |  |  |  |  |              *000 = 1:2   001 = 1:4
;   |  |  |  |  |               010 = 1:8   011 = 1:16
;   |  |  |  |  |               100 = 1:32  101 = 1:64
;   |  |  |  |  |               110 = 1:128 111 = 1:256
;   |  |  |  |  |______________ PSA (prescaler assign)
;   |  |  |  |                  1 = to WDT
;   |  |  |  |                 *0 = to Timer0
;   |  |  |  |_________________ TOSE (Timer0 edge select)
;   |  |  |                     0 = increment on low-to-high
;   |  |  |                    *1 = increment in high-to-low
;   |  |  |____________________ TOCS (TMR0 clock source)
;   |  |                       *0 = internal clock
;   |  |                        1 = RA4/TOCKI bit source
;   |  |_______________________ INTEDG (Edge select)
;   |                           0 = falling edge
;   |                          *1 = raising edge
;   |__________________________ RBPU (Pullup enable)
;                               0 = enabled
;                              *1 = dissabled
	movlw	b'11010000'  ; set up timer/counter
	movwf	OPTION_REG
	Bank0
	return
;============================================================
;                    receive character
;============================================================
; Receive a single character through the serial port.
; Assumes: 4800 baud, 8 data bits, no parity, 1 stop bit.
; Recieving line is Port A, line 0
rcvTTY:
	movlw	0x08		; Counter for 8 bits
	movwf	bitCount
; The start of character transmission is signaled by
; the sender by setting the line low
startBit:
	btfsc	PORTA,0		; Test for low on line
	goto	startBit	; Go if not low
;=========================
;  offset to data bit
;=========================
; At this point the receiver has found the falling
; edge of the start bit. It must now wait one and
; one-half the baud rate to synchronize in the center
; of the sender's first data bit, as follows:
;     |<========= falling edge of START bit
;     |     |<========== center of start bit
;     |     |           |<====== center of data bit
;     |-----------|-----|
;_____             ___________            __________
;     |           |           |          |        <== SIGNAL
;      -----------             ---------- 
;     |<-- 208 -->|<104>| <====== ms. for 4800 baud
; 
; Clock start count for one-half bit  = 255 - 104 = 151
; Clock start count for one full bit =  255 - 208 = 47
; One clock cycle is added for the movwf intruction: 
;    clkHalf = 152 (for one-half bit countdown)
;    clkFull = 48 (for one full bit countdown)
	movlw	halfBaud   	; Skip one-half bit
	movwf	TMR0		; Initialize tmr0 and start count
	bcf		INTCON,2   	; Clear overflow flag
;============================
;         start bit
;============================
wait1:
	btfss	INTCON,2   	; Timer count overflow?
	goto	wait1		; No, keep waiting
; At this point we are at the center of the start bit
	btfsc	PORTA,0		; Check to see it is still low
	goto	startBit	; No, it is high. False start
; At this point the clock is at the center of the start
; bit. The first data bit must be read one full baud
; period later
	movlw	fullBaud  	; One full bit delay	
	movwf	TMR0		; Start timer
	bcf		INTCON,2   	; clear tmr0 overflow flag
wait2:
	btfss	INTCON,2 	; End of one full baud period?
	goto	wait2		; Wait if not end of period
; Timer is now at the center of the first/next data bit
; Timer must be reset immediately so that code will not
; lose synchornization with sender 
	movlw	fullBaud	; Skip to next data bit	
	movwf	TMR0		; Restart timer
	bcf		INTCON,2   	; Reset overflow flag
; Now the data bit can be read and and stored
	movf	PORTA,w		; Read port B
	movwf	tempData	; Store in temporary variable
	rrf		tempData,f	; Rotate bit 0 into carry flag
	rrf		rcvData,f   ; Rotate carry flag into storage
						; register high-order bit
	decfsz	bitCount,f	; End of data?
	goto	wait2		; Continue until 8 bits received
;============================
;         stop bit
;============================
stopWait:
	btfss	INTCON,2   	; Test time
	goto	stopWait	; Wait
	return				; Exit

;============================================================
;                      send character
;============================================================
; Procedure to send one character through the RS232 line.
; Assumes: 2400 baud, 8 data bits, no parity, one stop bit
; Sending line is Port B, line 0
; ON ENTRY:
;         variable sendData holds character to send
sendTTY:
	movlw	0x08		; Init bit counter
	movwf	bitCount	
	bcf		PORTB,0		; Low for start bit
	movlw	fullBaud	; For one baud space
	movwf	TMR0		; Start timer
	bcf		INTCON,2  	; Clear timer flag
start2snd:
	btfss	INTCON,2   	; Full baud done?
	goto	start2snd	; No
	movlw	fullBaud	; Reset for one full bit period
	movwf	TMR0		; Start timer
	bcf		INTCON,2   	; Clear flag
; At this point the start bit has been sent
; Data follows
sendOut:
	rrf		sendData,f  ; Rotate bit into carry
	bcf		PORTB,0		; Assume data bit is 0
	btfsc	STATUS,c   	; Test if carry set
	bsf		PORTB,0 	; Change bit to 1 if clear
; Hold bit for 1 baud period
timeBit:
	btfss	INTCON,2   	; Wait for baud period to end
	goto	timeBit		; Loop if not yet
	movlw	fullBaud	; Reset timer
	movwf	TMR0		; Start timer
	bcf		INTCON,2   	; Clear flag
; Test for last bit
	decfsz	bitCount,f	; Count this bit
	goto	sendOut		; Continue if not last bit
; Done. Send stop bit
	bsf		PORTB,0		; High for stop bit
stopBit:
	btfss	INTCON,2   	; Timer done?
	goto	stopBit		; No
; Set port B line 0 high back again
	bsf		PORTB,0
	call	delay_5		; And hold 
	return

	end

