; File name: Tmr0Counter.asm
; Date: April 30, 2006
; Author: Julio Sanchez
; Processor: 16F84A
;
; Reference: SevenSeg Circuit and Board
;
; Description:
; Test program for the Timer0 counter. The program counts
; the number of presses of the pushbutton switch on port
; RA4/TOCKI and display the count on a seven segment LED.
; Switch is wired active low.
;
; Switches used in __config directive:
;   _CP_ON      	Code protection ON/OFF     
; * _CP_OFF      
; * _PWRTE_ON     Power-up timer ON/OFF
;   _PWRTE_OFF     
;   _WDT_ON       Watchdog timer ON/OFF  
; * _WDT_OFF       
;   _LP_OSC       Low power crystal occilator
; * _XT_OSC       External parallel resonator/crystal ocillator  
;   _HS_OSC       High speed crystal resonator (8 to 10 MHz)
;                Resonator: Murate Erie CSA8.00MG = 8 MHz   
;   _RC_OSC       Resistor/capacitor ocillator (simplest, 20% error)
; |
; |_____ * indicates setup values
;
;=========================
; setup and configuration
;=========================
	processor 16f84A
	include	  <p16f84A.inc>
	__config  _XT_OSC & _WDT_OFF & _PWRTE_ON & _CP_OFF
;
;=====================================================
;                 constant definitions
;              (per circuit wiring diagram)
;=====================================================
#define Pb_sw  4	; Port A line 4 to push button switch
;
;============================
;      local variables
;============================
	cblock	0x0c		; Start of block
	J				; counter J
	K				; counter K
	endc
;============================================================
;                           program
;============================================================
	org	0	  ; start at address 0
	goto	main
;
; Space for interrupt handlers
	org		0x08

main:
; Clear the timer and the watchdog
	clrf	TMR0
	clrwdt
; Set up the OPTION regiser bit map
	movlw	b'10111000'
;   7  6  5  4  3  2  1  0 <= OPTION bits
;   |  |  |  |  |  |__|__|_____ PS2-PS0 (prescaler bits)
;   |  |  |  |  |               Values for Timer0
;   |  |  |  |  |              *000 = 1:2   001 = 1:4
;   |  |  |  |  |               010 = 1:8   011 = 1:16
;   |  |  |  |  |               100 = 1:32  101 = 1:64
;   |  |  |  |  |               110 = 1:128 *111 = 1:256
;   |  |  |  |  |______________ PSA (prescaler assign)
;   |  |  |  |                 *1 = to WDT
;   |  |  |  |                  0 = to Timer0
;   |  |  |  |_________________ TOSE (Timer0 edge select)
;   |  |  |                     0 = increment on low-to-high
;   |  |  |                    *1 = increment in high-to-low
;   |  |  |____________________ TOCS (TMR0 clock source)
;   |  |                        0 = internal clock
;   |  |                       *1 = RA4/TOCKI bit source
;   |  |_______________________ INTEDG (Edge select)
;   |                          *0 = falling edge
;   |__________________________ RBPU (Pullup enable)
;                               0 = enabled
;                              *1 = dissabled
	option
; Setup ports
	movlw	0x00		; Set port B to output
	tris	PORTB
	clrf	PORTB		; All port B to 0
; Port A. Five low-order lines set for for input
	movlw	B'00011111'	; w = 00011111 binary
	tris	PORTA		; port A (lines 0 to 4) to input
;=================================
; Check value in TMR0 and display
;=================================
; Every press of the pushbutton switch connected to line
; RA4/TOCKI adds one to the value in the TMR0 register.
; Loop checks this value, adjusts to the range 0 to 15
; and displays the result in the seven-segment LED on
; port B
checkTmr0:
	movf	TMR0,w	; Timer register to w
; Elimate four high order bits
	andlw	b'00001111' ; Mask off high bits
; At this point the w register contains a 4-bit value
; in the range 0 to 0xf. Use this value (in w) to 
; obtain seven-segment display code
	call	segment
	movwf	PORTB		; Display switch bits
	goto	checkTmr0	; Endless loop

;================================
;  routine to returns 7-segment
;             codes
;================================
segment:
		addwf	PCL,f	; PCL is program counter latch
		retlw	0x3f	; 0 code
		retlw	0x06	; 1
		retlw	0x5b	; 2
		retlw	0x4f	; 3
		retlw	0x66	; 4
		retlw	0x6d	; 5
		retlw	0x7d	; 6
		retlw	0x07	; 7
		retlw	0x7f	; 8
		retlw	0x6f	; 9
		retlw	0x77	; A
		retlw	0x7c	; B
		retlw	0x39	; C
		retlw	0x5b	; D
		retlw	0x79	; E
		retlw	0x71	; F
		retlw	0x7f	; Just in case all on
		

		end	

